################################################################
#
#   Assembly Line - Company
#
################################################################

from __future__ import division
from math import sqrt
from datetime import date, time, datetime, timedelta
from random import random
from albow.dialogs import alert, ask
from game import game
from factory import Factory
from part import create_material
from product import ProductHistory, Crate
from imath import iround, iceil, prob_round, sig_round
from timing import period_for_date
from game import game
import config, settings
from settings import frames_per_second
from stats import Stats, StatsHolder
from format import format_dollars
from machines import all_machine_types, initial_machine_types
from bonus import create_random_bonus

#------------------------------------------------------------------------

debug_retailer = False
debug_customer = False

#------------------------------------------------------------------------

class MaterialStock(object):
	#  material    Material
	#  sheet_size  (width, height) sorted
	#  quantity    int
	#  average_cost  float
	
	quantity = 0
	average_cost = 0.0
	
	def __init__(self, material, size):
		self.material = material
		self.sheet_size = size
	
	def sheet_cost(self):
		return self.material.sheet_cost(self.sheet_size)

#------------------------------------------------------------------------

class ReportSelection(object):

	kind = None
	year = None
	month = None
	
	def period_range(self):
		k = self.kind
		if k == 'total':
			return 0, game.company.current_period() + 1
		elif k == 'last3':
			p2 = game.company.current_period()
			p1 = max(0, p2 - 3)
			n = p2 - p1
			return p1, n
		elif k == 'current':
			return game.company.current_period(), 1
		else:
			d = date(self.year, self.month, 1)
			p = period_for_date(d)
			return max(0, p), 1

#------------------------------------------------------------------------

class Transaction(object):
	#  transaction_id   int
	#  category         string
	#  description      string
	#  amount           int
	#  product_key      string
	#  factory_id       int
	#  quantity         int
	
	transaction_id = None
	category = None
	description = None
	amount = None
	product_key = None
	factory_id = None
	quantity = None

	def __init__(self, **kwds):
		for name in kwds:
			if hasattr(self, name):
				setattr(self, name, kwds[name])
			else:
				raise TypeError("%s.__init__: Keyword %s not allowed" % (
					self.__class__.__name__, name))
	
#------------------------------------------------------------------------

class CompanyStats(Stats):
	
	material_purchase = 0
	freight_cost = 0
	factory_rental = 0
	factory_running_cost = 0
	
	product_sale = 0
	material_sale = 0

	machinery_purchase = 0
	research_spending = 0
	machinery_sale = 0
	
	salary = 0
	bonus = 0
	
	def calc_expenditure(self):
		return (self.material_purchase
			+ self.factory_rental
			+ self.factory_running_cost
			+ self.freight_cost)
	
	def calc_income(self):
		return (self.product_sale
			+ self.material_sale)
	
	def calc_outstanding_capital(self):
		return (self.machinery_purchase
			+ self.research_spending
			- self.machinery_sale)
	
	def calc_balance_of_earnings(self):
		return self.income - self.expenditure
	
	def calc_profit(self):
		return self.balance_of_earnings - self.outstanding_capital
	
	def calc_net_cash_flow(self):
		return self.profit - self.salary - self.bonus
	
	expenditure = property(calc_expenditure)
	income = property(calc_income)
	outstanding_capital = property(calc_outstanding_capital)
	balance_of_earnings = property(calc_balance_of_earnings)
	profit = property(calc_profit)
	net_cash_flow = property(calc_net_cash_flow)

#------------------------------------------------------------------------

class Player(object):
	#  cash                int
	#  payments_suspended  boolean
	#  salary              int
	#  salary_raise_offer  int

	cash = 0
	payments_suspended = False
	salary_raise_offer = 0
	
	def __init__(self):
		self.init_salary()
	
	def init_salary(self):
		self.salary = config.starting_salary
	
	def __setstate__(self, d):
		self.__dict__.update(d)
		if 'salary' not in d:
			self.init_salary()

#------------------------------------------------------------------------

class Company(StatsHolder):
	#
	#  factories                [Factory]
	#  cash                     int
	#  machines_in_storage      {MachineType: [Machine]}
	#  machines_in_storage_list [(MachineType, [Machine])]
	#  material_stock           {(rgba, sorted_size): MaterialStock}
	#  material_stock_list      [MaterialStock]
	#  current_factory          Factory
	#  transactions             [(description, amount)]
	#  product_history          {key: ProductHistory}
	#  selected_material_stock  int or None
	#  selected_machine_store   int or None
	#  selected_product_no      int or None
	#  period_base              int
	#  datetime                 datetime
	#  frames                   int
	#  report_selection         ReportSelection
	#  time_since_retailer_visit int # game minutes
	#  time_since_customer_visit int # game minutes
	#  stats                    list[period] of CompanyStats
	#  total                    CompanyStats
	#  developed_products       [string]
	#  products_being_developed [string]
	#  rent_due                 int
	#  available_machine_types  [MachineType]
	#  machine_under_development MachineType
	#  months_to_next_machine   int
	#  months_to_machine_rumour int
	#  current_bonus            Bonus
	#  player                   Player
	
	current_factory = None
	selected_material_stock = None
	selected_machine_store = None
	selected_product_no = None
	time_since_retailer_visit = 0
	time_since_customer_visit = 0
	rent_due = 0
	machine_under_development = None
	months_to_next_machine = 0
	months_to_machine_rumour = 0
	current_bonus = None

	stats_class = CompanyStats

	def __init__(self):
		StatsHolder.__init__(self)
		self.factories = []
		self.cash = config.initial_cash
		self.machines_in_storage = {}
		self.machines_in_storage_list = []
		self.material_stock = {}
		self.material_stock_list = []
		self.transactions = []
		self.product_history = {}
		self.new_factory()
		self.init_date()
		self.init_report_selection()
		self.stats = []
		self.init_products()
		self.init_machine_types()
		self.init_player()
		self.next_bonus()
	
	def init_date(self):
		td = date.today()
		year = td.year
		month = td.month
		self.period_base = year * 12 + (month - 1)
		self.datetime = datetime(year, month, 1)
		self.frames = 0
	
	def init_report_selection(self):
		sel = ReportSelection()
		sel.kind = 'total'
		date = self.base_date()
		sel.year = date.year
		sel.month = date.month
		self.report_selection = sel
	
	def init_products(self):
		ap = game.all_products
		n = len(ap)
		d = min(config.num_products_initially_developed, n)
		u = min(config.num_products_initially_under_development, n - d)
		self.developed_products = [p.key for p in ap[:d]]
		self.products_being_developed = [p.key for p in ap[d:d+u]]
		#print "Company: developed_products =", self.developed_products ###
		#print "Company: products_being_developed =", self.products_being_developed ###

	def init_machine_types(self):
		self.available_machine_types = initial_machine_types[:]
		self.start_developing_next_machine()
		self.check_for_next_machine_rumour()
	
	def init_player(self):
		self.player = Player()
		self.pending_payments = []

	def __setstate__(self, d):
		self.__dict__.update(d)
		if 'material_stock' not in d:
			self.material_stock = {}
		if 'machines_in_storage_list' not in d:
			self.machines_in_storage_list = []
		if 'material_stock_list' not in d:
			self.material_stock_list = self.material_stock.items()
		if 'product_history' not in d:
			self.product_history = {}
		if 'datetime' not in d:
			self.init_date()
		if 'report_selection' not in d:
			self.init_report_selection()
		if 'stats' not in d:
			StatsHolder.__init__(self)
		if 'num_developed_products' not in d:
			self.init_products()
		if 'available_machine_types' not in d:
			self.init_machine_types()
		if not settings.record_transactions:
			del self.transactions[:]
		if 'current_bonus' not in d:
			self.next_bonus()
		if 'player' not in d:
			self.init_player()
	
	def num_developed_products(self):
		return len(self.developed_products)
	
	def num_products_under_development(self):
		return len(self.products_being_developed)
	
	def developed_product_no(self, n):
		return game.product_with_key(self.developed_products[n])
	
	def product_under_development_no(self, n):
		return game.product_with_key(self.products_being_developed[n])
	
	def product_is_developed(self, product):
		return product and product.key in self.developed_products
	
	def product_is_under_development(self, product):
		return product and product.key in self.products_being_developed
	
	def get_date(self):
		return self.datetime.date()
	
	date = property(get_date)
	
	def get_time(self):
		return self.datetime.time()
	
	time = property(get_time)
	
	def base_date(self):
		return self.period_start_date(0)
	
	def period_start_date(self, p):
		y, m = divmod(self.period_base + p, 12)
		return date(y, m + 1, 1)
	
	def base_year(self):
		return self.period_base // 12
	
	def num_years(self):
		return self.date.year - self.base_year() + 1
	
	def current_period(self):
		return period_for_date(self.date)

	def stats_for_previous_period(self, periods_ago):
		cp = self.current_period()
		return self.stats_for_period(cp - periods_ago)
	
	def get_profit(self):
		return self.income - self.expenditure
	
	profit = property(get_profit)
	
	def get_balance_of_earnings(self):
		return self.income - self.adjusted_expenditure
	
	balance_of_earnings = property(get_balance_of_earnings)
	
	def get_adjusted_expenditure(self):
		return self.expenditure - self.capital_outlay
	
	adjusted_expenditure = property(get_adjusted_expenditure)
	
	def new_factory(self):
		f = Factory()
		self.factories.append(f)
		self.select_factory(f)
	
	def select_factory(self, f):
		self.current_factory = f
	
	def close_factory(self):
		factory = self.current_factory
		title = factory.title
		num_machines = factory.num_machines()
		query = "Close factory '%s' containing %s machines?" % (title, num_machines)
		if ask(query, ["OK", "Cancel"]) == "OK":
			machines, crates, parts = factory.clear_factory()
			self.remove_factory(factory)
			num_crates = len(crates)
			num_products = 0
			for product, crate in crates:
				num_products += crate.num_items
				self.put_crate_into_stock(product, crate)
			value = 0
			num_parts = len(parts)
			for part in parts:
				value += part.raw_material_value()
			amount = self.sell_parts_for_scrap(num_parts, value)
			mess = "Factory '%s' closed.\n\n" % title
			if num_machines:
				mess += "\n\n%s machines put into storage." % num_machines
			if num_crates:
				mess += "\n\n%s crates containing %s products put into stock." \
					% num_crates
			if num_parts:
				mess += "\n\n%s parts sold as scrap for %s." % (num_parts,
					format_dollars(amount))
			alert(mess)
	
	def remove_factory(self, factory):
		factories = self.factories
		i = factories.index(factory)
		del factories[i]
		if not factories:
			self.new_factory()
		if factory is self.current_factory:
			self.select_factory(factories[min(i, len(factories) - 1)])
	
	def num_machines_in_storage(self, machine_type):
		store = self.machines_in_storage.get(machine_type)
		if store:
			return len(store)
		else:
			return 0
	
	def selected_machines_in_storage(self):
		n = self.selected_machine_store
		if n is not None:
			return self.machines_in_storage_list[n]

	def put_machine_into_storage(self, machine):
		key = machine.__class__
		store = self.machines_in_storage.get(key)
		if not store:
			store = []
			self.machines_in_storage[key] = store
			self.machines_in_storage_list.append((key, store))
		store.append(machine)
		print "Store now contains", len(store), key.__name__ ###

	def get_machine_from_storage(self, machine_type, match = None):
		store = self.machines_in_storage.get(machine_type)
		if store:
			if match:
				for machine in store:
					if match(machine):
						print "Got a matching", machine_type.__name__, "from store" ###
						return machine
			else:
				print "Getting a", machine_type.__name__, "from store" ###
				return store.pop()
	
	def get_material_sheet_from_stock(self, material, size, order_quantity, factory):
		#print "Getting a sheet of", material.formula(), size, "from stock" ###
		stock = self.find_material_stock(material, size)
		if not stock.quantity > 0 and order_quantity > 0:
			self.purchase_stock(stock, order_quantity, factory)
		if stock.quantity > 0:
			sheet = create_material(material, size)
			stock.quantity -= 1
			factory.add_stats(material_purchase = stock.average_cost)
			#print "There are now", stock.quantity, "sheets left in stock" ###
			return sheet
		#else:
			#print "There are no sheets left in stock" ###
	
	def find_material_stock(self, material, size):
		sorted_size = list(size)
		sorted_size.sort()
		sorted_size = tuple(sorted_size)
		key = (material.rgba(), sorted_size)
		stock = self.material_stock.get(key)
		if not stock:
			stock = MaterialStock(material, sorted_size)
			self.material_stock[key] = stock
			self.material_stock_list.append(stock)
		return stock
	
	def purchase_stock(self, stock, quantity, factory):
		cost = iceil(quantity * stock.sheet_cost())
		material = stock.material
		#print "Company.purchase_stock: stock.sheet_size =", stock.sheet_size ###
		size = "%sx%s" % stock.sheet_size
		formula = material.formula()
		desc = "%s %s" % (formula, size)
		if cost <= self.cash:
			self.spend_cash('material_purchase', desc, cost,
				factory_id = factory.id, quantity = quantity)
			old_qty = stock.quantity
			new_qty = old_qty + quantity
			stock.average_cost = (stock.average_cost * old_qty + cost) / new_qty
			stock.quantity = new_qty
		else:
			game.alert("Not enough cash to purchase %s %s sheets of %s "
				"(cost $%s, available funds $%s)" % (
					quantity, size, formula, cost, self.cash))

	def spend_cash(self, category, description, amount, **kwds):
		assert isinstance(amount, (int, long))
		if amount > self.cash:
			return False
		self.add_transaction(category, description, -amount, **kwds)
		self.cash -= amount
		self.record_amount(category, amount)
		#print "Spent: $%s on %s '%s' leaving $%s" % (amount, category, description, self.cash) ###
		return True

	def earn_cash(self, category, description, amount, **kwds):
		assert isinstance(amount, (int, long))
		self.add_transaction(category, description, amount, **kwds)
		self.cash += amount
		self.record_amount(category, amount)
		#print "Earned: $%s from %s '%s' giving $%s" % (amount, category, description, self.cash) ###

	def add_transaction(self, category, description, amount, **kwds):
		if settings.record_transactions:
			transactions = self.transactions
			id = len(transactions) + 1
			tran = Transaction(
				transaction_id = id,
				category = category,
				description = description,
				amount = amount,
				**kwds)
			self.transactions.append(tran)
	
	def record_amount(self, category, amount):
		v = {category: amount}
		self.stats_for_month(self.date).add_values(v)
		self.total.add_values(v)

	def begin_frame(self):
		if self.rent_due:
			if not self.pay_factory_rental():
				return
		n = settings.game_speeds[game.speed]
		while n:
			n -= 1
			for factory in self.factories:
				factory.begin_frame()
			f = self.frames + 1
			if f < frames_per_second:
				self.frames = f
			else:
				self.one_second_task()
				self.frames = 0
	
	def one_second_task(self):
		game.update_pending = True
		m = settings.game_minutes_per_second
		old_dt = self.datetime
		new_dt = old_dt + timedelta(minutes = m)
		if new_dt.day <> old_dt.day:
			self.daily_task()
		if new_dt.month <> old_dt.month:
			self.monthly_task()
		self.datetime = new_dt
		self.check_for_retailer_visit(m)
		self.check_for_customer_visit(m)
	
	def daily_task(self):
		self.pay_factory_running_costs()
		self.pay_research_costs()
		self.decay_product_reputations()
	
	def monthly_task(self):
		print "Company: Performing monthly task on", self.date ###
		self.calculate_factory_rental()
		self.perform_machine_development()
		self.check_for_next_machine_rumour()
		self.pay_salary()
		self.check_for_bonus_achievement()
		self.attempt_pending_payments()
		self.check_for_salary_raise()
		self.pay_factory_rental()

	def calculate_factory_rental(self):
		rental = config.factory_rental
		n = len(self.factories)
		self.rent_due += n * config.factory_rental
		for factory in self.factories:
			factory.add_stats(factory_rental = rental)
	
	def perform_machine_development(self):
		if self.machine_under_development:
			self.months_to_next_machine -= 1
			if self.months_to_next_machine <= 0:
				self.finish_developing_machine()
				self.start_developing_next_machine()
	
	def finish_developing_machine(self):
		mtype = self.machine_under_development
		self.available_machine_types.append(mtype)
		self.machine_under_development = None
		game.alert("New machine available: %s" % mtype.title)
	
	def start_developing_next_machine(self):
		for mtype in all_machine_types:
			if mtype not in self.available_machine_types:
				self.machine_under_development = mtype
				self.months_to_next_machine = mtype.months_to_develop
				self.months_to_machine_rumour = mtype.months_to_develop // 2
				return
	
	def check_for_next_machine_rumour(self):
		mtype = self.machine_under_development
		if mtype:
			if self.months_to_machine_rumour == 0:
				game.alert("%s machine rumoured to be under development." % mtype.title)
			self.months_to_machine_rumour -= 1
	
	def add_available_machine(self, mtype):
		available = self.available_machine_types
		if mtype not in available:
			available.append(mtype)

	def pay_factory_rental(self):
		amount = self.rent_due
		if amount <= 0:
			return True
		else:
			if self.spend_cash('factory_rental', "", amount):
				self.rent_due -= amount
				return True
			else:
				game.speed = 0
				alert("Game paused because there is not enough cash to pay\n\n"
					"factory rental of %s." % format_dollars(amount))
				return False
	
	def pay_factory_running_costs(self):
		for factory in self.factories:
			if factory.running:
				monthly_cost = factory.running_cost()
				#print "Company: Factory %s monthly running cost = %s" % (
				#	factory.id, monthly_cost) ###
				if monthly_cost:
					cost = iround(monthly_cost / 30)
					factory.add_stats(factory_running_cost = cost)
					if not self.spend_cash('factory_running_cost', "", cost,
							factory_id = factory.id):
						factory.stop_with_alert(
							"Factory stopped because there is not enough cash to\n"
							"pay running costs of %s." % format_dollars(cost))
	
	def pay_research_costs(self):
		finished = []
		for key in self.products_being_developed:
			product = game.product_with_key(key)
			history = self.find_product_history(product)
			budget = history.research_budget
			cost = prob_round(budget / 30)
			if cost:
				#print "Company: Research cost of %s = %s" % (product.title, cost) ###
				if self.spend_cash('research_spending', "", cost, product_key = product.key):
					#print "   Recording research spending" ###
					rate = product.development_rate(budget)
					research_done = rate / 30
					#print "   Research done = %.5f" % research_done ###
					history.research_spending += cost
					history.research_done += research_done
					if history.research_done >= 1.0:
						#print "   Research finished" ###
						finished.append(product)
				else:
					history.research_budget = 0
					game.alert("Research budget of %s reduced to 0 because there is\n"
						"not enough cash to pay it." % product.title)
		for product in finished:
			self.finish_developing_product(product)
			self.start_developing_new_product()
	
	def finish_developing_product(self, product):
		#print "Company: Finishing development of", product.title ###
		key = product.key
		self.products_being_developed.remove(key)
		self.developed_products.append(key)
		game.alert("Development of %s is complete." % product.title)
	
	def start_developing_new_product(self):
		developed = self.developed_products
		developing = self.products_being_developed
		for product in game.all_products:
			key = product.key
			if key not in developed and key not in developing:
				developing.append(key)
				game.alert("New product ready for development: %s" % product.title)
				return

	def check_for_retailer_visit(self, minutes_elapsed):
		n = self.time_since_retailer_visit + minutes_elapsed
		d = config.retailer_restocking_interval
		while n >= d:
			n -= d
			self.simulate_retailers()
		self.time_since_retailer_visit = n
	
	def check_for_customer_visit(self, minutes_elapsed):
		n = self.time_since_customer_visit + minutes_elapsed
		d = config.customer_visit_interval
		while n >= d:
			n -= d
			self.simulate_customers()
		self.time_since_customer_visit = n
	
	def find_product_history(self, product):
		history = self.product_history.get(product.key)
		if not history:
			history = ProductHistory(product)
			self.product_history[product.key] = history
		return history
	
	def put_products_into_stock(self, product, errors, factory):
		quality_sum = 0
		num_items = len(errors)
		for error in errors:
			quality = max(0.0, 1.0 - error / 0.05)
			quality_sum += quality
		average_quality = quality_sum / num_items
		crate = Crate(num_items, average_quality, factory)
		self.put_crate_into_stock(product, crate)
	
	def put_crate_into_stock(self, product, crate):
		history = self.find_product_history(product)
		#print "Putting a crate of %s x %s with avg quality %s into stock" % (
		#	num_items, product.title, crate.average_quality)
		history.crates.append(crate)
	
	def simulate_retailers(self):
		purchases_per_prospective_customer = (config.customer_visit_rate
			* config.retailer_restocking_interval / config.customer_visit_interval)
		for key, history in self.product_history.items():
			product = game.product_with_key(key)
			crates_for_sale = history.crates
			if not crates_for_sale:
				continue
			if debug_retailer:
				print "Company: Retailer considering purchase of %s" % product.title
			if not history.is_for_sale:
				if debug_retailer:
					print "   Not for sale"
				continue
			estimated_demand = iceil(purchases_per_prospective_customer
				* history.prospective_customers)
			current_stock = history.retailer_stock_level()
			if debug_retailer:
				print "   Estimated demand =", estimated_demand
				print "   Current stock =", current_stock
			num_sold = 0
			crates_sold = 0
			num_crates_for_sale = len(crates_for_sale)
			selling_price = history.current_price
			area = product.area()
			tot_revenue = 0
			tot_freight_cost = 0
			while estimated_demand > current_stock and crates_sold < num_crates_for_sale:
				crate = crates_for_sale[crates_sold]
				num_items = crate.num_items
				if debug_retailer:
					print "   Selling crate of %s to retailer" % num_items
				crates_sold += 1
				current_stock += num_items
				num_sold += num_items
				revenue = iceil(num_items * selling_price)
				bulk = num_items * area
				freight_cost = config.base_freight_cost + iceil(bulk * config.freight_rate)
				tot_revenue += revenue
				tot_freight_cost += freight_cost
				factory = crate.factory
				if factory:
					factory.add_stats(items_sold = num_items, product_sale = revenue,
						freight_cost = freight_cost)
			if debug_retailer and current_stock < estimated_demand:
				print "   Not enough crates available" ###
			if crates_sold > 0:
				if self.sell_products(product, history, num_sold, selling_price,
						tot_revenue, tot_freight_cost):
					if debug_retailer:
						print "   Transferring %s crates to retailer" % crates_sold
					history.retailer_stock.extend(crates_for_sale[:crates_sold])
					del crates_for_sale[:crates_sold]
	
	def simulate_customers(self):
		product_history = self.product_history
		for key in product_history:
			history = product_history[key]
			product = game.product_with_key(key)
			#if history.retailer_stock:
			self.simulate_customer_purchases(product, history)
			self.check_for_market_saturation(product, history)
			self.simulate_customer_communication(product, history)

	def simulate_customer_purchases(self, product, history):
		date = self.date
		num_customers = prob_round(config.customer_visit_rate
			* history.prospective_customers)	
		if num_customers > 0:
			if debug_customer:
				print "Company: %s customers considering purchase of %s" % (num_customers, product.title)
			#r = history.current_price / product.price
			##q = abs(r - 1.0) ** config.expected_quality_power
			#a = config.expected_quality_coefficient
			#q = a * abs(r - 1.0) ** 2
			#if r > 1.0:
			#	expected_quality = 1.0 + q
			#else:
			#	expected_quality = 1.0 - q
			#if debug_customer:
			#	print "   Expected quality = %.2f" % expected_quality
			#---
			reputation = history.biased_reputation_ratio()
			if debug_customer:
					print "   Biased reputation = %.2f" % reputation
			#buying_probability = reputation / expected_quality
			expected_price = reputation * product.price
			selling_price = history.current_price
			if selling_price > expected_price:
				overpricing = selling_price / expected_price - 1.0
				max_overpricing = config.max_overpricing
				loss_probability = min(1.0, overpricing / max_overpricing)
				if debug_customer:
					print "   Loss probability = %.2f" % loss_probability
				lost_customers = prob_round(loss_probability * num_customers)
				num_customers -= lost_customers
				history.lose_customers(date, lost_customers)
				if 0: #lost_customers: ###
					print "Lost customers:", lost_customers ###
					print "   rec =", history.positive_reputation ###
					print "   not rec =", history.negative_reputation ###
					print "   reputation =", reputation ###
					print "   expected_price =", expected_price ###
					print "   selling_price =", selling_price ###
					print "   overpricing =", overpricing ###
					print "   loss_probability =", loss_probability ###
					print "   num_customers =", num_customers ###
			stock = history.retailer_stock
			i = 0
			while num_customers:
				while stock and stock[0].num_items < 1:
					if debug_customer:
						print "   Discarding empty crate"
					del stock[0]
				if not stock:
					if debug_customer:
						print "   No more crates in stock"
					break
				crate = stock[0]
				#if random() < buying_probability:
				if 1:
					if debug_customer:
						print "   Customer %s: Buying" % i
					actual_quality = crate.average_quality
					satisfaction_probability = actual_quality # / expected_quality
					satisfied = random() < satisfaction_probability
					if debug_customer:
						print "   Satisfaction probability = %.2f" % satisfaction_probability
						print "   Satisfied = %s" % satisfied
					crate.num_items -= 1
					assert crate.num_items >= 0
					history.add_customer(date, satisfied)
				#else:
				#	if debug_customer:
				#		print "   Customer %s: Not buying" % i
				#	history.lose_customer()
				i += 1
				num_customers -= 1
			if debug_customer:
				print "   %s customers going away empty-handed" % num_customers
			history.bored_customers(num_customers)

	def check_for_market_saturation(self, product, history):
		if not history.market_saturation_reported:
			total = history.total
			ac = (history.prospective_customers + history.unprospective_customers
				+ history.potential_market)
			tc = history.total_market
			if ac / tc < config.market_saturation_warning_level:
				history.market_saturation_reported = True
				game.alert("Marketing reports that the %s market is "
					"becoming saturated. We need new products to sell."
					% product.title)

	def simulate_customer_communication(self, product, history):
		total = history.total
		pr = history.positive_reputation
		nr = history.negative_reputation
		if debug_customer:
			print "Company: Reputation +%8.3f -%8.3f for %s" % (pr, nr, product.title)
		pc = history.prospective_customers
		uc = history.unprospective_customers
		mp = history.potential_market
		mt = history.total_market
		if mt == 0:
			return
		rc = config.customer_conversion_rate
		rd = config.unprospective_customer_decay_rate
		#tr = pr + nr
		#if tr:
		#	fr = 1.0 - nr / tr
		#else:
		#	fr = 1.0
		rp = config.advertising_target_purchase_rate
		fm = config.advertising_target_market_fraction
		tv = config.customer_visit_interval / 1440 # days
		#a = (rp * tv) / (fm * mt)
		#amax = (rp * tv) / mp
		a = (rp * tv) / max(fm * mt, mp)
		#fp = history.biased_reputation_ratio()
		#fn = 1.0 - fp
		#nmp = prob_round(pr * rc * mp / mt + fp * a * mp)
		#nmu = prob_round(nr * rc * mp / mt + fn * a * mp)
		nmp = prob_round(pr * rc * mp / mt + a * mp)
		nmu = prob_round(nr * rc * mp / mt)
		num = prob_round(rd * uc)
		#w = config.customer_hysteresis_factor
		#fup = pr - w * nr
		#fpu = nr - w * pr
		#print "Company: fup = %.3f fpu = %.3f" % (fup, fpu) ###
		#if fup > 0.0:
		#	xup = fup * rc * uc / mt
		#	#print "Company: xup = %.3f" % xup
		#	npu = -prob_round(xup)
		#elif fpu > 0.0:
		#	xpu = fpu * rc * pc / mt
		#	#print "Company: xpu = %.3f" % xpu
		#	npu = prob_round(xpu)
		#else:
		#	npu = 0
		if history.convert_potential_to_prospective_customers(nmp):
			if debug_customer:
				print "Company: %s potential to prospective customers for %s" % (nmp, product.title)
		if history.convert_potential_to_unprospective_customers(nmu - num):
			if debug_customer:
				print "Company: %s potential to unprospective customers for %s" % (nmu, product.title)
		#history.lost_customers(num)
		#if npu:
		#	if history.convert_prospective_to_unprospective_customers(npu):
		#		if debug_customer:
		#			print "Company: %s prospective to unprospective customers for %s" % (npu, product.title)
		
	def decay_product_reputations(self):
		r = config.reputation_decay_factor
		product_history = self.product_history
		for key in product_history:
			history = product_history[key]
			history.positive_reputation *= r
			history.negative_reputation *= r

	def sell_products(self, product, history, num_sold, selling_price,
			revenue, freight_cost):
		#history = self.find_product_history(product)
		#print "Company: Selling %s x %s for $%s each" % (
		#	num_sold, product.title, selling_price) ###
		#revenue = iceil(num_sold * selling_price)
		#bulk = num_sold * product.area()
		#freight_cost = config.base_freight_cost + iceil(bulk * config.freight_rate)
		#print "   Revenue = $%s" % revenue ###
		#print "   Freight cost = $%s" % freight_cost ###
		key = product.key
		if not self.spend_cash('freight_cost', "", freight_cost,
				product_key = key, quantity = num_sold):
			history.is_for_sale = False
			game.alert("Sales of %s suspended because there is not enough cash "
				"to pay freight costs." % product.title)
			return False
		history.add_stats(
			self.date,
			number_sold = num_sold,
			revenue = revenue,
			freight_cost = freight_cost)
		title = product.title
		desc = "%s Freight cost" % title
		if revenue <> 0:
			desc = "%s sales %s @ $%s" % (title, num_sold, selling_price)
			self.earn_cash('product_sale', desc, revenue,
				product_key = key, quantity = num_sold)
		return True

	def sell_parts_for_scrap(self, num_items, total_value):
		amount = iround(config.scrap_value_factor * total_value)
		self.earn_cash('material_sale', "Scrap", amount, quantity = num_items)
		return amount
	
	def pay_salary(self):
		self.queue_payment('salary', "", self.player.salary)

	def next_bonus(self):
		if not self.current_bonus:
			self.current_bonus = create_random_bonus(self)
	
	def announce_bonus(self):
		bonus = self.current_bonus
		if bonus:
			mess = ("New Target Announced:\n\n"
				"\n\n"
				"%s of %s\n\n"
				"\n\n"
				"Bonus: %s") % (
					bonus.title_text(),
					bonus.target_text(),
					bonus.reward_text())
			game.alert(mess)

	def check_for_bonus_achievement(self):
		bonus = self.current_bonus
		if bonus and bonus.achieved(self):
			title = bonus.title_text()
			target = bonus.target_text()
			reward = bonus.reward_text()
			game.alert("Target Achieved:\n\n"
				"\n\n"
				"%s of %s\n\n"
				"\n\n"
				"Bonus: %s" % (title, target, reward))
			desc = "%s %s" % (title, target)
			self.queue_payment('bonus', desc, bonus.reward)
			self.current_bonus = None
			self.next_bonus()
			self.announce_bonus()
	
	def queue_payment(self, category, desc, amount):
		#print "Company: Queueing payment:", category, desc, amount ###
		self.pending_payments.append([category, desc, amount])

	def attempt_pending_payments(self):
		if self.player.payments_suspended:
			print "Company: Checking whether to resume payments" ###
			if not self.salary_is_affordable():
				return
		queue = self.pending_payments
		while queue:
			pending = queue[0]
			category, desc, amount = pending
			min_cash = amount * config.payout_cash_surplus_factor
			payment = min(amount, self.cash - min_cash)
			if payment <= 0:
				self.unable_to_pay()
				return
			self.pay_player(category, desc, payment)
			pending[2] -= payment
			if pending[2] > 0:
				break
			queue.pop(0)
		self.player.payments_suspended = False

	def pay_player(self, category, desc, amount):
		self.spend_cash(category, desc, amount)
		self.player.cash += amount
	
	def unable_to_pay(self):
		player = self.player
		if not player.payments_suspended:
			player.payments_suspended = True
			game.alert("We're sorry, but the company does not have enough surplus "
			  "cash to pay all of your salary and bonuses right now. We'll pay you "
			  "the rest as soon as the company's financial position improves.")

	def get_salary_owed(self):
		return self.get_amount_owed('salary')

	def get_bonuses_owed(self):
		return self.get_amount_owed('bonus')
	
	def get_amount_owed(self, category):
		total = 0
		for pending in self.pending_payments:
			if pending[0] == category:
				total += pending[2]
		return total
	
	def salary_is_affordable(self):
		surplus = self.affordable_salary()
		print "Company: affordable salary =", surplus ###
		return surplus >= self.player.salary
	
	def affordable_salary(self):
		p = self.current_period()
		recent_profit = 0
		for stats in self.stats[-3:]:
			recent_profit += stats.profit
		sf = config.salary_surplus_factor
		surplus = recent_profit * sf / 3
		return surplus

	def check_for_salary_raise(self):
		surplus = self.affordable_salary()
		rf = config.salary_raise_factor
		player = self.player
		new_offer = sig_round(player.salary * rf, 2)
		current_offer = player.salary_raise_offer
		print "Company: check_for_salary_raise: new offer", new_offer, \
			"surplus", surplus, "current_offer", current_offer ###
		if surplus >= new_offer and new_offer > current_offer:
			player.salary_raise_offer = new_offer
			self.announce_salary_raise_offer()

	def announce_salary_raise_offer(self):
		offer = self.player.salary_raise_offer
		if offer:
			answer = ask("Company offers to raise your salary to %s."
					% format_dollars(12 * offer),
				responses = ["Accept", "Decline"])
			if answer == "Accept":
				self.accept_salary_raise_offer()
			else:
				game.alert("If you change your mind, click the Accept button "
					"beside the Salary Offer notice on the Company page.")
	
	def accept_salary_raise_offer(self):
		player = self.player
		offer = player.salary_raise_offer
		if offer:
			player.salary = offer
			player.salary_raise_offer = 0
			game.alert("Salary raised to %s per year." % format_dollars(12 * offer))
