################################################################
#
#   Assembly Line - Company page
#
################################################################

from __future__ import division
from albow.widget import Widget
from albow.controls import Label, Row, Column, Grid, Frame, ValueView, Button, \
	IntEditor, IntField, BoolField
from albow.dialogs import alert, Dialog
from albow.palette_view import PaletteView
from albow.table_view import TableView, TableColumn as TC
from albow.resource import get_font
from screen import ALScreen
from decoration import Lineflow, PostitTabPanel
from digital_datetime import DigitalDateView
from game_speed_control import GameSpeedControl
from game import game
from format import format_dollars, format_money, format_bool
from imath import iceil, iround
from product_view import ProductTableBase
from research_subpage import ResearchSubpage
from bonus_view import BonusView, SalaryView
import config

#-----------------------------------------------------------------------

heading_font = get_font(15, "VeraBd.ttf")
subheading_font = get_font(12, "VeraBd.ttf")
table_header_font = get_font(12, "VeraBd.ttf")
table_data_font = get_font(12, "Vera.ttf")

#-----------------------------------------------------------------------

class CompanyPage(ALScreen):

	bg_color = (157, 75, 7)
	margin = 16
	right_pane_tab_font = get_font(15, "Vera.ttf")

	def __init__(self, shell, height):
		ALScreen.__init__(self, shell)
		m = self.margin
		date_view = DigitalDateView()
		financial_summary = self.create_financial_summary()
		personal = self.create_personal_view()
		speed_control = GameSpeedControl()
		left_pane = Column([
			date_view,
			financial_summary,
			personal,
		], align = 'l')
		right_pane = PostitTabPanel(
			tab_font = self.right_pane_tab_font,
			width = self.width - left_pane.width - 3 * m,
			anchor = 'lrtb')
		right_column = Column([
			right_pane,
			speed_control,
		], align = 'l', anchor = 'lrtb', expand = right_pane,
			height = self.height - 2 * m)
		stock = self.create_stock_view()
		equipment = self.create_equipment_view()
		products = self.create_product_view()
		research = ResearchSubpage()
		right_pane.add_page("Material", stock)
		right_pane.add_page("Equipment", equipment)
		right_pane.add_page("Products", products)
		right_pane.add_page("Research", research)
		contents = Row([left_pane, right_column],
			align = 't', spacing = m, anchor = 'ltrb')
		contents.topleft = (m, m)
		self.add(contents)
	
	def create_financial_summary(self):
		def H(text):
			return Label(text, font = subheading_font)
		def L(text):
			return Label(text, font = table_data_font)
		rows = [
			[Label("Financial Summary", font = heading_font), None],
			[L("Cash"), CashView()],
			[H("This Month"), None],
			[L("Income"), MonthView('income')],
			[L("Expenditure"), MonthView('expenditure')],
			[L("Balance of Earnings"), MonthView('balance_of_earnings')],
			[H("Lifetime"), None],
			[L("Balance of Earnings"), TotalView('balance_of_earnings')],
			[L("Net Capital Outlay"), TotalView('outstanding_capital')],
			[L("Profit"), TotalView('profit')],
		]
		return Lineflow(Grid(rows), margin = 10)
	
	def create_personal_view(self):
		bonus = BonusView()
		salary = SalaryView()
		contents = Column([bonus, salary], align = 'l')
		return Lineflow(contents)
	
	def create_stock_view(self):
		label = Label("Materials in Stock", font = heading_font)
		table = StockView()
		controls = Button("Sell",
			enable = stock_is_selected, action = sell_stock)
		contents = Column([
			label,
			table,
			controls,
		], align = 'l', expand = 1)
		return Lineflow(contents)

	def create_equipment_view(self):
		label = Label("Machinery in Storage", font = heading_font)
		table = EquipmentView()
		controls = Button("Sell",
			enable = equipment_is_selected, action = sell_equipment)
		contents = Column([
			label,
			table,
			controls,
		], align = 'l', expand = 1)
		return Lineflow(contents)

	def create_product_view(self):
		label = Label("Products in Stock", font = heading_font)
		table = ProductView()
		controls = Row([
			Button("Settings...",
				enable = selected_product_is_developed,
				action = edit_developed_product),
			Button("Scrap",
				enable = selected_product_is_developed,
				action = sell_product_for_scrap),
		])
		contents = Column([
			label,
			table,
			controls,
		], align = 'l', expand = 1)
		return Lineflow(contents)

#-----------------------------------------------------------------------

money_width = format_dollars(999999999)

class MoneyView(ValueView):

	font = table_data_font
	format_value = staticmethod(format_dollars)
	align = 'r'
	
	def __init__(self, **kwds):
		ValueView.__init__(self, width = money_width, **kwds)

class TotalView(MoneyView):

	def __init__(self, attr, **kwds):
		MoneyView.__init__(self, attribute = attr, **kwds)
	
	def get_value(self):
		return getattr(game.company.total, self.attribute)

class MonthView(MoneyView):
	#  periods_ago   int

	def __init__(self, attr, periods_ago = 0, **kwds):
		self.periods_ago = periods_ago
		MoneyView.__init__(self, attribute = attr, **kwds)
	
	def get_value(self):
		s = game.company.stats_for_previous_period(self.periods_ago)
		return getattr(s, self.attribute)

class CashView(MoneyView):

	def get_value(self):
		return game.company.cash

#def IncomeView():
#	return CompanyMonthView('income')
#
#def ExpenditureView():
#	return CompanyMonthView('adjusted_expenditure')
#
#def BOEView():
#	return CompanyMonthView('balance_of_earnings')
#
#def OCView():
#	return CompanyMonthView('capital_outlay')

def ProfitView():
	return CompanyMonthView('profit')

#-----------------------------------------------------------------------

class StockView(TableView):

	header_font = table_header_font
	font = table_data_font

	columns = [
		TC("Composition", 120, 'l'),
		TC("Size",         50, 'l', "%sx%s"),
		TC("Sheets",       50, 'r'),
		TC("Cost",        100, 'r', format_dollars),
	]
	
	def num_rows(self):
		return len(game.company.material_stock_list)

	def row_is_selected(self, n):
		return n == game.company.selected_material_stock

	def click_row(self, n, e):
		game.company.selected_material_stock = n

	def row_data(self, n):
		stock = game.company.material_stock_list[n]
		type = stock.material.formula()
		size = stock.sheet_size
		qty = stock.quantity
		cost = iceil(qty * stock.sheet_cost())
		return (type, size, qty, cost)

#-----------------------------------------------------------------------

class EquipmentView(TableView):

	header_font = table_header_font
	font = table_data_font

	columns = [
		TC("Type", 180, 'l'),
		TC("Qty",   50, 'r'),
		TC("Cost", 100, 'r', format_dollars),
	]
	
	def num_rows(self):
		return len(game.company.machines_in_storage_list)
	
	def row_is_selected(self, n):
		return n == game.company.selected_machine_store
	
	def click_row(self, n, e):
		game.company.selected_machine_store = n
	
	def row_data(self, n):
		machine_type, store = game.company.machines_in_storage_list[n]
		qty = len(store)
		cost = 0
		for machine in store:
			cost += machine.cost
		return (
			machine_type.title,
			qty,
			cost,
		)

#-----------------------------------------------------------------------

class ProductView(ProductTableBase, TableView):

	header_font = table_header_font
	font = table_data_font
	
	selectable = True

	columns = [
		TC("Product",       150, 'l'),
		TC("Crates",         50, 'r'),
		TC("Items",          50, 'r'),
		TC("Price",          50, 'r', format_dollars),
		TC("For Sale",       70, 'r', format_bool),
	]
	
	def row_data(self, n):
		p, h = self.product_and_history_no(n)
		return (
			p.title,
			h.crates_in_stock(),
			h.number_in_stock(),
			h.current_price,
			h.is_for_sale,
		)

#-----------------------------------------------------------------------

def stock_is_selected():
	return game.company.selected_material_stock is not None

def sell_stock():
	company = game.company
	stocks = company.material_stock_list
	n = company.selected_material_stock
	if n is not None and n < len(stocks):
		stock = stocks[n]
		material = stock.material
		size = "%sx%s" % stock.sheet_size
		description = "%s %s" % (material.formula(), size)
		title = "Sell Material: " + description
		price = config.resale_price_factor * stock.sheet_cost()
		dlog = SellDialog(title, stock.quantity, "sheets", price, "per sheet")
		if dlog.present():
			qty = dlog.quantity
			stock.quantity -= qty
			amount = iceil(qty * price)
			company.earn_cash('material_sale', description, amount, quantity = qty)
			alert("%s sheets of %s sold for %s." % (
				qty, description, format_dollars(amount)))

#-----------------------------------------------------------------------

def selected_product_is_developed():
	return game.company.product_is_developed(game.selected_product)

def edit_developed_product():
	product = game.selected_product
	history = game.company.find_product_history(product)
	name_field = Label(product.title)
	price_field = IntField(min = 0, width = "999999")
	for_sale_field = BoolField()
	grid = Grid([
		[Label("Product"), name_field],
		[Label("Selling Price"),
			Row([Row([Label("$"), price_field], spacing = 5), Label("each")])],
		[Label("For Sale"), for_sale_field],
	])
	dlog = Dialog(grid, responses = ["OK", "Cancel"])
	price_field.value = history.current_price
	for_sale_field.value = history.is_for_sale
	if dlog.present() == "OK":
		#print "change_product_price: OK" ###
		history.current_price = price_field.value
		history.is_for_sale = for_sale_field.value

def sell_product_for_scrap():
	product = game.selected_product
	history = game.company.find_product_history(product)
	crates = history.crates
	num_crates = len(crates)
	num_items = 0
	quality_sum = 0.0
	for crate in crates:
		n = crate.num_items
		num_items += n
		quality_sum += n * crate.average_quality
	if num_items:
		average_quality = quality_sum / num_items
	else:
		average_quality = 0.0
	#print "sell_product_for_scrap: average_quality =", average_quality ###
	f = config.scrap_value_factor
	part_cost = product.part.raw_material_cost()
	if num_crates:
		avg_crate_cost = part_cost * num_items / num_crates
	else:
		avg_crate_cost = 0.0
	avg_crate_price = f * average_quality * avg_crate_cost
	title = "Sell for Scrap: " + product.title
	dlog = SellDialog(title, num_crates, "crates", avg_crate_price, "each (average)")
	if dlog.present():
		num_crates_to_sell = dlog.quantity
		if num_crates_to_sell:
			crates_to_sell = crates[:num_crates_to_sell]
			del crates[:num_crates_to_sell]
			num_items = 0
			total = 0
			for crate in crates_to_sell:
				n = crate.num_items
				num_items += n
				price = n * f * crate.average_quality * part_cost
				total += price
				#print "...selling a crate of", crate.num_items, "for", price ###
			stats = history.stats_for_current_period()
			history.number_trashed += num_items
			amount = iround(total)
			game.company.earn_cash('material_sale', "Scrap", amount,
				quantity = num_items, product_key = product.key)
			alert("%s x %s sold as scrap for %s." % (num_items, product.title,
				format_dollars(amount)))

#-----------------------------------------------------------------------

def equipment_is_selected():
	return game.company.selected_machine_store is not None

def sell_equipment():
	company = game.company
	entry = company.selected_machines_in_storage()
	if entry:
		mtype, store = entry
		title = mtype.title
		price = config.resale_price_factor * mtype.cost
		desc = "Sell %s Machines" % title
		dlog = SellDialog(desc, len(store), "", price, "each")
		if dlog.present():
			qty = dlog.quantity
			del store[:qty]
			amount = iceil(qty * price)
			company.earn_cash('machinery_sale', title, amount, quantity = qty)
			alert("%s %s machines sold for %s." % (qty, title, format_dollars(amount)))

#-----------------------------------------------------------------------

class SellDialog(Dialog):

	def __init__(self, title, max_qty, qty_unit, price, price_unit):
		Widget.__init__(self)
		self.price = price
		self.quantity = max_qty
		header = Label(title)
		qty_field = IntEditor(min = 0, max = max_qty, attribute = 'quantity', model = self)
		grid = Grid([
			[Label("Quantity"), Row([qty_field, Label(qty_unit)])],
			[Label("Price"), Label("$%.2f %s" % (price, price_unit))],
			[Label("Total"), ValueView(format_value = format_dollars, get_value = self.total)],
		])
		buttons = Row([Button("OK", action = self.ok),
			Button("Cancel", action = self.cancel)])
		contents = Column([header, grid, buttons], margin = self.margin)
		self.add(contents)
		self.shrink_wrap()
		
	def total(self):
		return iceil(self.price * self.quantity)
		
