################################################################
#
#   Assembly Line - Controls
#
################################################################

from pygame import draw
from pygame.display import get_surface
from pygame.locals import K_LEFT, K_RIGHT
from albow.dialogs import Modal, Dialog
from albow.controls import Label, Button, Column, ValueView, ImageButton
from albow.palette_view import PaletteView
from albow.resource import get_font, get_image
from albow.utils import blit_in_rect
from albow.theme import ThemeProperty
from dropdown import DropdownListField
from format import month_names
from game import game
from product_view import ProductPaletteBase

#-----------------------------------------------------------------------

class Multichoice(PaletteView):

	highlight_color = ThemeProperty('highlight_color')

	#unfocused_sel_color = (128, 128, 128)
	highlight_style = 'frame'
	align = 'c'
	tab_stop = True

	def __init__(self, cell_size, values, attribute, model, **kwds):
		PaletteView.__init__(self, cell_size, 1, len(values), **kwds)
		self.values = values
		self.attribute = attribute
		self.model = model

	def num_items(self):
		return len(self.values)
	
	def item_is_selected(self, n):
		return self.get_value() == self.values[n]
	
	def click_item(self, n, e):
		if self.tab_stop:
			self.focus()
		self.set_value(self.values[n])
	
	def get_value(self):
		return getattr(self.model, self.attribute)
	
	def set_value(self, x):
		setattr(self.model, self.attribute, x)
	
	def draw(self, surf):
		if self.has_focus():
			surf.fill(self.highlight_color)
		PaletteView.draw(self, surf)

	def key_down(self, e):
		k = e.key
		if k == K_LEFT:
			self.change_value(-1)
		elif k == K_RIGHT:
			self.change_value(1)
		else:
			PaletteView.key_down(self, e)
	
	def change_value(self, d):
		values = self.values
		if values:
			n = len(values)
			value = self.get_value()
			try:
				i = values.index(value)
			except ValueError:
				if d < 0:
					i = 0
				else:
					i = n - 1
			else:
				i = max(0, min(n - 1, i + d))
			self.set_value(values[i])

#-----------------------------------------------------------------------

class ImageMultichoice(Multichoice):

	highlight_style = 'fill'
	sel_color = (255, 192, 19)
	margin = 5

	def __init__(self, images, values, **kwds):
		image0 = images[0]
		w, h = image0.get_size()
		d = 2 * self.predict(kwds, 'margin')
		cell_size = w + d, h + d
		Multichoice.__init__(self, cell_size, values, **kwds)
		self.images = images

	def draw_item(self, surf, n, rect):
		image = self.images[n]
		blit_in_rect(surf, image, rect, self.align, self.margin)
	
	def draw_prehighlight(self, surf, i, rect):
		color = self.sel_color
		surf.fill(color, rect)

#-----------------------------------------------------------------------

class TextMultichoice(Multichoice):

	margin = 5
	#sel_color = (128, 128, 128)
	sel_color = (64, 64, 64)

	def __init__(self, values, labels = None, **kwds):
		if not labels:
			labels = map(str, values)
		font = self.predict_font(kwds)
		widths = [w for (w, h) in map(font.size, labels)]
		d = 2 * self.predict(kwds, 'margin')
		cw = max(widths) + d
		ch = font.get_linesize() + d
		Multichoice.__init__(self, (cw, ch), values, **kwds)
		self.labels = labels

	def draw_item(self, surf, n, rect):
		buf = self.font.render(self.labels[n], True, self.fg_color)
		blit_in_rect(surf, buf, rect, self.align, self.margin)

	def draw_prehighlight(self, surf, i, rect):
		d = -self.font.get_descent()
		m = self.margin + d
		color = self.sel_color
		x, y = rect.midtop
		y += d
		pts = [(x - m, y - m), (x + m, y - m), (x, y)]
		draw.polygon(surf, color, pts)
		x, y = rect.midbottom
		y -= d + 1
		pts = [(x - m, y + m), (x + m, y + m), (x, y)]
		draw.polygon(surf, color, pts)
	
#-----------------------------------------------------------------------

reset_image = get_image("reset_button.png")

class MachineEditor(Dialog):

	title_font = get_font(18, "VeraBd.ttf")
	
	def __init__(self, factory, machine):
		Dialog.__init__(self)
		self.factory = factory
		self.machine = machine
		title = Label("Configure " + machine.title, font = self.title_font)
		inner = self.create_contents(machine)
		title_and_inner = Column([title, inner], spacing = 16, align = 'c')
		ok_button = Button("OK", self.ok)
		contents = Column([title_and_inner, ok_button], align = 'r')
		contents.topleft = (16, 16)
		self.add(contents)
		self.shrink_wrap()
		if machine.resettable:
			reset_button = ImageButton(reset_image, action = self.reset)
			reset_button.bottomleft = (16, self.height - 16)
			self.add(reset_button)

	def reset(self):
		self.machine.reset(self.factory)
	
#-----------------------------------------------------------------------

class ProductKeyField(ValueView):

	border_width = 1
	margin = 3

	def __init__(self, width = 200, **kwds):
		ValueView.__init__(self, width = width, **kwds)
	
	def mouse_down(self, e):
		dlog = ProductKeyDialog(width = self.width,
			attribute = self.attribute, model = self.model,
			bg_color = self.inherited('bg_color'))
		p = self.local_to_global((0, 0))
		dlog.topleft = p
		sw, sh = get_surface().get_size()
		if dlog.bottom > sh:
			dlog.bottom = sh
		if dlog.top < 0:
			dlog.top = 0
		dlog.present(centered = False)
	
	def format_value(self, key):
		if key:
			return game.product_with_key(key).title
		else:
			return ""

#-----------------------------------------------------------------------

class ProductKeySelector(ProductPaletteBase, PaletteView):

	cell_width = 150
	lmargin = 5
	highlight_style = 'reverse'
	fg_color = (0, 0, 0)
	bg_color = (255, 255, 255)
	#sel_color = (62, 188, 254)
	#sel_color = (255, 255, 255)
	border_width = 1
	attribute = None
	model = None

	def __init__(self, width = 200, height = 300, **kwds):
		font = self.predict_font(kwds)
		cw = width
		ch = font.get_linesize() + 4
		nr = height // ch
		PaletteView.__init__(self, (cw, ch), nr, 1, scrolling = True, **kwds)
		self.scroll_to_value(self.get_value())
	
	def scroll_to_value(self, key):
		product = game.product_with_key(key)
		if product:
			try:
				n = game.company.developed_products.index(product)
			except ValueError:
				pass
			else:
				self.scroll_to_item(n)
	
	def item_is_selected(self, n):
		product = self.product_no(n)
		return product and product.key == self.get_value()
	
	def click_item(self, n, e):
		product = self.product_no(n)
		if product:
			self.set_value(self.product_no(n).key)
	
	def draw_item(self, surf, n, rect):
		product = self.product_no(n)
		if product:
			buf = self.font.render(product.title, True, self.fg_color)
			r = buf.get_rect()
			r.midleft = rect.left + self.lmargin, rect.centery
			surf.blit(buf, r)
	
	def get_value(self):
		return getattr(self.model, self.attribute)
	
	def set_value(self, x):
		setattr(self.model, self.attribute, x)

#-----------------------------------------------------------------------

class ProductKeyDialog(Modal, ProductKeySelector):

	def click_item(self, n, e):
		ProductKeySelector.click_item(self, n, e)
		self.dismiss(True)

#-----------------------------------------------------------------------

class MonthField(DropdownListField):

	default_width = "May"

	def num_items(self):
		return 12
	
	def value_for_item_no(self, n):
		return n + 1
	
	def item_no_for_value(self, x):
		if x is not None:
			return x - 1
		else:
			return None

	def format_value(self, x):
		if x is not None:
			return month_names[x - 1]
		else:
			return ""

#-----------------------------------------------------------------------

class YearField(DropdownListField):

	default_width = "9999"

	def num_items(self):
		return game.company.num_years()
	
	def value_for_item_no(self, n):
		return game.company.base_year() + n
	
	def item_no_for_value(self, x):
		if x is not None:
			return x - game.company.base_year()
		else:
			return None
