################################################################
#
#   Assembly Line - Factory page
#
################################################################

from pygame import Rect, draw
from pygame.locals import *
from albow import Cancel
from albow.widget import Widget
from albow.controls import Column, ImageButton
from albow.palette_view import PaletteView
from albow.resource import get_image, get_font
from albow.image_array import get_image_array
from albow.dialogs import alert, ask, Dialog
from screen import ALScreen
from transformation import Transformation, identity, add, sub
from settings import options
from factory import FactoryError
from game import game
from format import format_dollars
from machine_type_view import MachineTypePaletteBase
from factory_menu import FactoryMenu

class FactoryPage(ALScreen):

	bg_color = (128, 180, 204)
	
	def __init__(self, shell):
		ALScreen.__init__(self, shell)
		menu_button = ImageButton("factory_menu_button.png", size = (32, 32), action = self.factory_menu)
		self.menu_button = menu_button
		light = FactoryRunLight(self.game)
		start_button = ImageButton("start_button.png", size = (32, 32), action = self.start_factory)
		stop_button = ImageButton("stop_button.png", size = (32, 32), action = self.stop_factory)
		palette = FactoryPalette(self.game)
		sweep = FactoryButton("sweep_tool.png", self.sweep_factory)
		machine = CurrentMachineView()
		controls = Column([
			menu_button, light, start_button, stop_button, palette, sweep, machine
		], spacing = 0)
		controls.right = self.right
		r = Rect(0, 0, controls.left, self.height - 24)
		factory_view = FactoryView(r, self.game)
		self.add(factory_view)
		self.add(controls)
		factory_view.focus()
	
	def factory_menu(self):
		dlog = FactoryMenu()
		dlog.topright = self.menu_button.local_to_global((-2, 2))
		dlog.present(centered = False)
	
	def start_factory(self):
		try:
			factory = self.game.factory()
			if not factory.running:
				factory.check_ready()
				unpurchased = factory.unpurchased_machines()
				if unpurchased:
					self.purchase_machines(factory, unpurchased)
				factory.start()
		except Cancel:
			pass
		except FactoryError, e:
			alert(str(e))
	
	def purchase_machines(self, factory, machines):
		if not options.dbg_no_costs:
			self.pay_for_machines(factory, machines)
		for machine in machines:
			machine.purchased = True
	
	def pay_for_machines(self, factory, machines):
		cost = 0
		for machine in machines:
			cost += machine.cost
		company = self.game.company
		cash = company.cash
		if cost > cash:
			alert("Not enough cash to purchase machinery.\n"
				"Cost: $%s Available funds: $%s" % (cost, cash))
			raise Cancel
		if ask("Purchase new machinery for $%s?" % cost) <> "OK":
			raise Cancel
		company.spend_cash('machinery_purchase', "", cost,
			factory_id = factory.id)
		#company.capital_outlay += cost

	def stop_factory(self):
		self.game.factory().stop()
	
	def sweep_factory(self):
		n, price = game.factory().sweep()
		if n:
			alert("Swept up %s parts from the factory floor.\n\n"
				"Sold as scrap for %s." % (n, format_dollars(price)))
		else:
			alert("The factory is already clean.")
	
	def begin_frame(self):
		factory = game.factory()
		game.play_sounds_for_factory(factory)
		if factory.running:
			self.invalidate()

#-----------------------------------------------------------------------

class FactoryButton(ImageButton):

	def __init__(self, image_name, action):
		ImageButton.__init__(self,
			size = (32, 32),
			image = get_image(image_name),
			action = action)

#-----------------------------------------------------------------------

class FactoryRunLight(Widget):

	images = get_image_array("factory_run_light.png", 2, border = 2)
	
	def __init__(self, game):
		self.game = game
		r = self.images[0].get_rect()
		Widget.__init__(self, r)
	
	def draw(self, surf):
		image = self.images[self.game.factory().running]
		surf.blit(image, (0, 0))

#-----------------------------------------------------------------------

class FactoryPalette(PaletteView):

	cell_size = (32, 32)
	tools = ['move', 'add', 'remove', 'configure', 'inspect']
	highlight_style = 'fill'

	def __init__(self, game):
		self.game = game
		PaletteView.__init__(self, self.cell_size, self.num_items(), 1)
	
	def num_items(self):
		return len(self.tools)
	
	def item_is_selected(self, n):
		return self.game.factory_tool == self.tools[n]
	
	def click_item(self, n, e):
		self.game.factory_tool = self.tools[n]
	
	def draw_item(self, surf, n, rect):
		image_name = self.tools[n] + "_tool.png"
		image = get_image(image_name)
		r = image.get_rect()
		r.center = rect.center
		surf.blit(image, r)

#-----------------------------------------------------------------------

class CurrentMachineView(Widget):

	#border_color = (192, 128, 0)
	#border_width = 2

	def __init__(self):
		Widget.__init__(self, rect = Rect(0, 0, 32, 32))
	
	def draw(self, surf):
		frame = surf.get_rect()
		mtype = game.selected_machine_type
		if mtype:
			image = mtype.small_icon
			r = image.get_rect()
			r.center = frame.center
			surf.blit(image, r)
		else:
			r = Rect(0, 0, 28, 28)
			r.center = frame.center
			surf.fill(MachineMenu.bg_color, r)
	
	def mouse_down(self, e):
		menu = MachineMenu()
		p = self.local_to_global((0, 0))
		menu.topright = p
		menu.present(centered = False)

#-----------------------------------------------------------------------

class MachineMenu(Dialog):

	bg_color = (103, 177, 109)
	border_width = 1
	margin = 5
	
	click_outside_response = True

	def __init__(self):
		contents = MachineSelector(self.ok)
		Dialog.__init__(self, contents)

#-----------------------------------------------------------------------

class MachineSelector(MachineTypePaletteBase):

	highlight_style = 'fill'
	sel_color = (0, 0, 255)

	def __init__(self, action):
		machine_types = game.company.available_machine_types
		nc = 8
		nr = (len(machine_types) + nc - 1) // nc
		MachineTypePaletteBase.__init__(self, cell_size = (32, 32),
			nrows = nr, ncols = nc)
		self.action = action
	
	def click_item(self, n, e):
		MachineTypePaletteBase.click_item(self, n, e)
		self.action()
	
	def draw_item(self, surf, n, rect):
		image = self.machine_type_no(n).small_icon
		r = image.get_rect()
		r.center = rect.center
		surf.blit(image, r)

#-----------------------------------------------------------------------


class FactoryView(Widget):

	error_marker_image = get_image("factory_error_marker.png")
	title_font = get_font(15, "Vera.ttf")
	title_color = (0, 0, 255)
	hint_font = title_font
	hint_color = title_color
	rotation_hint = "Press R to rotate machine"

	gridline_color = (96, 140, 159)
	moving_mouse_position = (0, 0)
	#moving_machine_type = None
	moving_machine = None
	moving_delta = (0, 0)
	moving_rotation = 0
	#feedback_okay = False
	#feedback_rects = None
	
	def __init__(self, rect, game):
		Widget.__init__(self, rect)
		self.game = game
	
	def factory(self):
		return self.game.factory()
	
	def factory_running(self):
		return self.factory().running
	
	def view_to_global_point(self, p):
		return p

	def global_to_view_point(self, p):
		return p

	def draw(self, surf):
		factory = self.factory()
		if factory:
			self.draw_factory(surf, factory)
	
	def draw_factory(self, surf, factory):
		self.draw_grid_lines(surf, factory)
		factory.draw(surf, identity)
		self.draw_feedback(surf, factory)
		self.draw_error(surf, factory)
		self.draw_title(surf, factory)
	
	def draw_title(self, surf, factory):
		color = self.title_color
		buf = self.title_font.render(factory.title, True, color)
		r = buf.get_rect()
		r.topright = (self.width - 5, 5)
		surf.blit(buf, r)
	
	def draw_grid_lines(self, surf, factory):
		cw, ch = factory.grid_size
		wc, hc = factory.cell_size
		color = self.gridline_color
		w = cw * wc
		h = ch * hc
		for cx in xrange(cw + 1):
			x = cx * wc
			draw.line(surf, color, (x, 0), (x, h))
		for cy in xrange(ch + 1):
			y = cy * hc
			draw.line(surf, color, (0, y), (w, y))

	def draw_feedback(self, surf, factory):
		rects = None
		tool = self.game.factory_tool
		if tool == 'add' or tool == 'move':
			factory = self.factory()
			if not factory.running:
				okay, rects, rotatable = self.get_moving_feedback()
		if rects:
			if okay:
				color = (0, 255, 0)
			else:
				color = (255, 0, 0)
			for rl in rects:
				for r in rl:
					draw.rect(surf, color, r, 1)
			if rotatable:
				self.draw_rotation_hint(surf)
	
	def draw_rotation_hint(self, surf):
		frame = surf.get_rect()
		buf = self.hint_font.render(self.rotation_hint, True, self.hint_color)
		r = buf.get_rect()
		r.midbottom = (frame.centerx, frame.bottom - 8)
		surf.blit(buf, r)
	
	def draw_error(self, surf, factory):
		p = factory.error_position
		if p:
			q = self.global_to_view_point(p)
			image = self.error_marker_image
			r = image.get_rect()
			r.center = q
			surf.blit(image, r)
	
	def machine_to_add(self):
		mtype = self.game.selected_machine_type
		if mtype:
			machine = self.game.company.get_machine_from_storage(mtype)
			if not machine:
				machine = mtype()
			return machine

	def mouse_move(self, e):
		#print "FactoryPage.mouse_move" ###
		tool = self.game.factory_tool
		if tool == 'add':
			self.add_tool_move(e)
	
	def mouse_down(self, e):
		tool = self.game.factory_tool
		if tool == 'move':
			self.move_tool_click(e)
		elif tool == 'add':
			self.add_tool_click(e)
		elif tool == 'remove':
			self.remove_tool_click(e)
		elif tool == 'configure':
			self.configure_tool_click(e)
		elif tool == 'inspect':
			self.inspect_tool_click(e)
	
	def mouse_drag(self, e):
		tool = self.game.factory_tool
		if tool == 'move':
			self.move_tool_drag(e)
	
	def mouse_up(self, e):
		tool = self.game.factory_tool
		if tool == 'move':
			self.move_tool_release(e)
	
	def key_down(self, e):
		if not e.alt or e.cmd:
			if e.key == K_r:
				self.rotate_cmd(e)
				return
		Widget.key_down(self, e)
	
	def add_tool_move(self, e):
		factory = self.factory()
		if not factory.running:
			mtype = self.game.selected_machine_type
			if mtype:
				self.moving_mouse_position = self.view_to_global_point(e.local)
				self.moving_delta = (0, 0)
				self.invalidate()
				self.invalidate()
				return
	
	def add_tool_click(self, e):
		factory = self.factory()
		if not factory.running:
			machine = self.machine_to_add()
			if machine:
				self.moving_mouse_position = self.view_to_global_point(e.local)
				t = self.moving_transformation()
				#if factory.add_machine(machine, t):
				#	self.cancel_moving_feedback()
				#else:
				if not factory.add_machine(machine, t):
					if machine.purchased:
						self.game.company.put_machine.into_storage(machine)

	def move_tool_click(self, e):
		factory = self.factory()
		if not factory.running:
			p = self.view_to_global_point(e.local)
			machine = factory.find_machine_at(p)
			if machine:
				self.moving_machine = machine
				self.moving_mouse_position = p
				self.moving_delta = sub(machine.translation, p)
				self.moving_rotation = machine.rotation
				self.invalidate()
	
	def move_tool_drag(self, e):
		machine = self.moving_machine
		if machine:
			self.moving_mouse_position = self.view_to_global_point(e.local)
			#self.update_moving_feedback()
			self.invalidate()
	
	def get_moving_feedback(self):
		okay = False
		rects = None
		rotatable = False
		mtype = self.moving_machine_type()
		if mtype:
			rotatable = mtype.rotation_mask
			machine = self.moving_machine
			t = self.moving_transformation()
			if t:
				okay, rects, t = \
					self.factory().new_machine_type_position(mtype, t,
						ignore_machine = machine)
		return okay, rects, rotatable
	
	def moving_machine_type(self):
		tool = game.factory_tool
		if tool == 'add':
			return game.selected_machine_type
		elif tool == 'move':
			machine = self.moving_machine
			if machine:
				return machine.__class__
	
	def move_tool_release(self, e):
		machine = self.moving_machine
		if machine:
			okay, rects, rotatable = self.get_moving_feedback()
			if okay:
				self.moving_mouse_position = self.view_to_global_point(e.local)				
				old_t = machine.local_to_global_transformation()
				t = self.moving_transformation()
				factory = self.factory()
				factory.remove_machine(machine)
				if not factory.add_machine(machine, t):
					if not factory.add_machine(machine, old_t):
						# Shouldn't be possible, but just in case
						self.game.company.put_machine_into_storage(machine)
			self.moving_machine = None
	
	def moving_transformation(self):
		m = self.moving_mouse_position
		if m:
			p = add(self.moving_delta, m)
			r = self.moving_rotation & self.moving_machine_type().rotation_mask
			return Transformation(r, p)
	
	def remove_tool_click(self, e):
		factory = self.factory()
		if not factory.running:
			machine = self.find_clicked_machine(e)
			if machine:
				factory.remove_machine(machine)
				if machine.purchased:
					components = []
					machine.extract_components(components)
					company = game.company
					company.put_machine_into_storage(machine)
					for component in components:
						company.put_machine_into_storage(component)
	
	def configure_tool_click(self, e):
		if not self.factory_running():
			machine = self.find_clicked_machine(e, layer = 1)
			if machine:
				editor_class = machine.editor_class
				if editor_class:
					editor = editor_class(self.factory, machine)
					editor.present()
	
	def inspect_tool_click(self, e):
		machine = self.find_clicked_machine(e)
		if machine:
			self.factory().inspect(machine)
			self.invalidate()
	
	def find_clicked_machine(self, e, layer = None):
		p = self.view_to_global_point(e.local)
		return self.factory().find_machine_at(p, layer)

	def rotate_cmd(self, e):
		tool = self.game.factory_tool
		if tool == 'add' or tool == 'move':
			self.change_rotation()
	
	def change_rotation(self):
		self.moving_rotation = (self.moving_rotation + 1) % 4
		self.invalidate()
