################################################################
#
#   Assembly Line - Game
#
################################################################

# See bottom for imports

class Game(object):
	#  company                Company
	#  selected_machine_type  MachineType
	#  factory_tool           string
	#  all_products           [Product]
	#  selected_product       Product
	#  running                boolean
	#  update_pending         boolean
	#  speed                  int
	#  factory_sound_queues   {factory: {sound: 1}}
	
	company = None
	selected_machine_type = None
	factory_tool = 'select'
	save_dir = None
	save_name = None
	selected_product = None
	running = True
	update_pending = True
	speed = 1
	
	def init(self):
		all_products = load_products()
		self.all_products = all_products
		map = {}
		for product in all_products:
			map[product.key] = product
		self.products_by_key = map
		self.frames = settings.frames_per_second
		self.factory_sound_queues = {}

	def save_game_enabled(self):
		return self.game_in_progress()

	def resume_game_enabled(self):
		return self.game_in_progress()

	def game_in_progress(self):
		return self.company is not None
	
	def unpause(self):
		if self.speed < 1:
			self.speed = 1
	
	def new_game(self):
		from company import Company
		self.company = Company()
		self.save_name = None
	
	def load_game(self, path):
		f = open(path, "rb")
		self.check_save_magic(f, settings.app_magic)
		self.check_save_magic(f, settings.save_file_magic)
		p = Unpickler(f)
		file_version = p.load()
		app_version = p.load()
		self.check_save_version(file_version, app_version)
		company = p.load()
		f.close()
		self.company = company
		self.set_save_path(path)
	
	def check_save_magic(self, f, expected):
		magic = f.read(len(expected))
		if magic <> expected:
			raise EnvironmentError("Does not seem to be a saved game file.")
	
	def check_save_version(self, file_version, app_version):
		min_version = settings.save_min_version
		cur_version = settings.save_file_version
		if not min_version <= file_version <= cur_version:
			if file_version > cur_version:
				mess = "Can only be opened by version %s.%s.%s or later." % app_version
			else:
				v1, v2 = settings.save_version_ranges[file_version]
				vs1 = "%s.%s.%s" % v1
				vs2 = "%s.%s.%s" % v2
				mess = "Can only be opened by versions %s to %s." % (vs1, vs2)
			raise EnvironmentError(mess)
	
	def save_game(self, path):
		fd, tmp_path = mkstemp(dir = os.path.dirname(path))
		f = os.fdopen(fd, "wb")
		f.write(settings.app_magic)
		f.write(settings.save_file_magic)
		p = Pickler(f, 2)
		p.dump(settings.save_file_version)
		p.dump(settings.save_app_version)
		p.dump(self.company)
		f.close()
		os.rename(tmp_path, path)
		self.set_save_path(path)
	
	def set_save_path(self, path):
		self.save_dir, self.save_name = os.path.split(path)

	def factory(self):
		return self.company.current_factory
	
	def get_save_dir(self):
		return self.save_dir or settings.get_default_save_dir()
	
	def get_save_path(self):
		dir = self.get_save_dir()
		name = self.save_name or settings.save_file_default_name
		return os.path.join(dir, name)

	def begin_frame(self):
		#print "game.begin_frame", self.frames ###
		if self.speed:
			company = self.company
			if company:
				company.begin_frame()
	
	def factory_running(self):
		company = self.company
		if company:
			return company.current_factory.running
	
	def product_with_key(self, key):
		return self.products_by_key.get(key)
	
	def product_is_selected(self, product):
		return product is self.selected_product
	
	def product_no_is_selected(self, n):
		return self.selected_product is self.all_products[n]
	
	def select_product(self, product):
		self.selected_product = product
	
	def select_product_no(self, n):
		if n is not None:
			self.selected_product = self.all_products[n]
		else:
			self.selected_product = None
	
	def alert(self, mess):
		response = ask(mess, responses = ["Pause", "Slow", "Continue"],
			default = 1, cancel = 2)
		if response == "Pause":
			self.speed = 0
		elif response == "Slow":
			self.speed = 1
	
	def change_speed(self, d):
		s = self.speed + d
		if 0 <= s < len(settings.game_speeds):
			self.speed = s
	
	def queue_sound_for_factory(self, factory, sound):
		queues = self.factory_sound_queues
		queue = queues.get(factory)
		if not queue:
			queue = {}
			queues[factory] = queue
		queue[sound] = 1
	
	def play_sounds_for_factory(self, factory):
		queue = self.factory_sound_queues.get(factory)
		if queue:
			for sound in queue:
				sound.play()
	
	def flush_sound_queue(self):
		self.factory_sound_queues.clear()

#------------------------------------------------------------------------

game = Game()

#------------------------------------------------------------------------

import os
from datetime import date, timedelta
from tempfile import mkstemp
from cPickle import Pickler, Unpickler
from albow.dialogs import ask
import settings
from product import load_products
