################################################################
#
#   Assembly Line - Inspection page
#
################################################################

from pygame import Rect, draw
from albow.widget import Widget
from albow.controls import Label, Row, Column, Grid
from albow.palette_view import PaletteView
from screen import ALScreen
from product_view import ProductView
from game import game
from machines import QCStation
from material import draw_formula

class InspectionPage(ALScreen):

	bg_color = (192, 192, 192)
	
	def __init__(self, shell):
		ALScreen.__init__(self, shell)
		height = self.height - 24
		comparison = PartComparisonView(pixel_size = 9,
			get_part1 = self.get_inspected_part,
			get_part2 = self.get_reference_part)
		analysis = AnalysisView(comparison)
		contents = Row([comparison, analysis], align = 't')
		contents.topleft = (16, 16)
		self.add(contents)

	def get_inspected_part(self):
		machine = self.get_inspected_qc_station()
		if machine:
			return machine.last_rejected_part
	
	def get_reference_part(self):
		machine = self.get_inspected_qc_station()
		if machine:
			product = game.product_with_key(machine.product_key)
			if product:
				return product.part
	
	def get_inspected_qc_station(self):
		machine = game.factory().inspected_machine
		if isinstance(machine, QCStation):
			return machine

#------------------------------------------------------------------------

def AnalysisView(model):
	return Grid([
		[Label("Part"), ColorView('selected_color_1', model)],
		[Label("Specification"), ColorView('selected_color_2', model)],
	])

#------------------------------------------------------------------------

class ColorView(Widget):

	model = None
	attribute = None
	
	def __init__(self, attribute, model, **kwds):
		r = Rect(0, 0, 150, self.predict_font(kwds).get_linesize())
		Widget.__init__(self, r, attribute = attribute, model = model)
	
	def get_value(self):
		model = self.model
		if model:
			return getattr(model, self.attribute)
	
	def draw(self, surf):
		frame = self.get_margin_rect()
		r = Rect(frame.topleft, (frame.height, frame.height))
		c = self.get_value()
		if c and c[3]:
			surf.fill(c, r)
			draw_formula(surf, c, (r.right + 5, r.top), self.fg_color)

#------------------------------------------------------------------------

class PartComparisonView(Widget):

	bg_color = (0, 0, 0)
	selection = (-1, -1)

	def __init__(self, pixel_size, **kwds):
		self.pixel_size = pixel_size
		size = 60 * pixel_size
		r = Rect(0, 0, size, size)
		Widget.__init__(self, r, **kwds)
	
	def get_part1(self):
		raise NotImplementedError

	def get_part2(self):
		raise NotImplementedError
	
	def get_selected_color_1(self):
		return self.selected_color(0)
	
	def get_selected_color_2(self):
		return self.selected_color(1)
		
	selected_color_1 = property(get_selected_color_1)
	selected_color_2 = property(get_selected_color_2)
	
	def selected_color(self, i):
		pixels, rects, bounds = self.pixels_and_rects()
		sx, sy = self.selection
		r = rects[i]
		if r.left <= sx < r.right and r.top <= sy < r.bottom:
			return pixels[i].get_at((sx - r.left, sy - r.top))
	
	def mouse_down(self, e):
		x, y = e.local
		ps = self.pixel_size
		self.selection = (x // ps, y // ps)
		#print "PartComparisonView.mouse_down: selection =", self.selection ###
	
	def pixels_and_rects(self):
		pixels1 = pixels2 = None
		part1 = self.get_part1()
		part2 = self.get_part2()
		if part1:
			pixels1 = part1.pixels
			rect1 = pixels1.get_rect()
		else:
			rect1 = Rect(0, 0, 0, 0)
		if part2:
			pixels2 = part2.pixels
			rect2 = pixels2.get_rect()
		else:
			rect2 = Rect(0, 0, 0, 0)
		rect2.center = rect1.center
		bounds = rect1.union(rect2)
		bounds.topleft = (0, 0)
		rect1.center = rect2.center = bounds.center
		return (pixels1, pixels2), (rect1, rect2), bounds

	def draw(self, surf):
		#white = (255, 255, 255)
		red = (255, 0, 0)
		green = (0, 255, 0)
		sel_red = (255, 255, 0)
		sel_green = (0, 0, 255)
		bg = self.bg_color
		draw_line = draw.line
		draw_rect = draw.rect
		ps = self.pixel_size
		pw = ps - 1
		width1 = height1 = width2 = height2 = 0
		(pixels1, pixels2), (rect1, rect2), bounds = self.pixels_and_rects()
		hrange = xrange(bounds.width)	
		vrange = xrange(bounds.height)
		x1a, y1a = rect1.topleft
		x2a, y2a = rect2.topleft
		x1b, y1b = rect1.bottomright
		x2b, y2b = rect2.bottomright
		#print "PartComparisonView.draw:" ###
		#print "rect1 =", rect1 ###
		#print "rect2 =", rect2 ###
		#print "limits 1 =", x1a, y1a, x1b, y1b ###
		#print "limits 2 =", x2a, y2a, x2b, y2b ###
		sx, sy = self.selection
		for y in vrange:
			for x in hrange:
				sel = x == sx and y == sy
				a1 = a2 = 0
				c1 = c2 = None
				if x1a <= x < x1b and y1a <= y < y1b:
					c1 = pixels1.get_at((x - x1a, y - y1a))
					a1 = c1[3]
				if x2a <= x < x2b and y2a <= y < y2b:
					c2 = pixels2.get_at((x - x2a, y - y2a))
					a2 = c2[3]
				pix = Rect(x * ps, y * ps, ps, ps)
				bc = None
				if pixels1 and (a1 or a2):
					if c1 == c2:
						if sel:
							bc = sel_green
						else:
							bc = green
					else:
						if sel:
							bc = sel_red
						else:
							bc = red
					surf.fill(bc, pix)
					#draw_rect(surf, bc, pix, 1)
				if a1 or a2:
					if a1:
						fc = c1
						d = -4
					elif a2:
						fc = c2
						d = -6
					if fc[:3] == bc:
						r = pix.inflate(-2, -2)
						surf.fill(bg, r)
					pix.inflate_ip(d, d)
					surf.fill(fc, pix)
