################################################################
#
#   Assembly Line - Input Hopper
#
################################################################

from pygame import Rect
from albow.controls import Label, Row, Grid, IntEditor, \
	ValueView
from albow.resource import get_image, get_font, get_sound
from machine import MachineNotConfiguredError
from part import create_material
from material import MaterialType
from config import standard_sheet_sizes
from material_view import MaterialField
from controls import TextMultichoice, MachineEditor
from hopper import Hopper, SupplyRateField
from game import game

#------------------------------------------------------------------------

class InputHopperEditor(MachineEditor):

	def create_contents(self, machine):
		matview = MaterialField(machine, 'product_key', machine.material)
		rows = [
			[Label("Material"), matview],
			[Label("Sheet Width"), SheetSizeField('sheet_width', machine)],
			[Label("Sheet Height"), SheetSizeField('sheet_height', machine)],
			[Label("Supply Rate"), SupplyRateField(machine, "sheets")],
			[Label("Material Price"), PixelCostView(machine),],
			[Label("Stock Cost"), SheetCostView(machine),],
		]
		return Grid(rows)
	
#------------------------------------------------------------------------

def SheetSizeField(attribute, machine):
	return TextMultichoice(standard_sheet_sizes, margin = 2, border_width = 1,
		attribute = attribute, model = machine)

def PixelCostView(machine):
	return ValueView(format = "$%.2f per kilopixel",
		width = "$999999 per kilopixel",
		attribute = 'pixel_cost', model = machine)

def SheetCostView(machine):
	return ValueView(format = "$%.2f per sheet",
		width = "$999999 per sheet",
		attribute = 'sheet_cost', model = machine)

#------------------------------------------------------------------------

class InputHopper(Hopper):
	#  product_key     string
	#  material        MaterialType
	#  _sheet_width    int
	#  _sheet_height   int
	#  order_quantity  int

	title = "Supply Hopper"
	icon = get_image("input_hopper_icon.png")
	on_left = True
	image = get_image("input_hopper.png")
	cost = 1000
	editor_class = InputHopperEditor
	
	orgx, orgy = -32, -64
	
	local_bounding_rect = Rect(orgx, orgy, 64, 96)

	local_covered_rects = [
		[Rect(orgx, orgy, 64, 32)],
		[Rect(orgx, orgy, 64, 96)]
	]
	
	product_key = None
	_sheet_width = None
	_sheet_height = None
	order_quantity = 20
	
	def __init__(self):
		self.material = MaterialType()
	
	def __setstate__(self, d):
		if 'sheet_width' in d:
			d['_sheet_width'] = d.pop('sheet_width')
		if 'sheet_height' in d:
			d['_sheet_height'] = d.pop('sheet_height')
		self.__dict__.update(d)
		if 'material' not in d:
			self.material = MaterialType()
	
	def get_pixel_cost(self):
		return self.material.cost
	
	pixel_cost = property(get_pixel_cost)
	
	def get_sheet_width(self):
		return self._sheet_width
	
	def get_sheet_height(self):
		return self._sheet_height
	
	def set_sheet_width(self, x):
		if x < 10:
			self._sheet_height = max(10, self._sheet_height)
		self._sheet_width = x
	
	def set_sheet_height(self, x):
		if x < 10:
			self._sheet_width = max(10, self._sheet_width)
		self._sheet_height = x
	
	sheet_width = property(get_sheet_width, set_sheet_width)
	sheet_height = property(get_sheet_height, set_sheet_height)
	
	def get_sheet_size(self):
		return (self.sheet_width, self.sheet_height)
	
	sheet_size = property(get_sheet_size)
	
	def get_sheet_cost(self):
		return self.material.sheet_cost(self.sheet_size)
	
	sheet_cost = property(get_sheet_cost)
	
	def check_ready(self):
		Hopper.check_ready(self)
		self.material.check_ready()
		if self.sheet_width is None or self.sheet_height is None \
			or self.order_quantity is None:
				raise MachineNotConfiguredError

	def get_part(self, factory):
		company = game.company
		material = self.material
		size = self.sheet_size
		ord_qty = self.order_quantity
		part = company.get_material_sheet_from_stock(material, size, ord_qty, factory)
		if part:
			part.tag = 'raw material'
			return part
		else:
			factory.stop_with_alert(
				"Factory stopped because a supply hopper is out of stock.")
	
