################################################################
#
#   Assembly Line - Punch
#
################################################################

from math import pi
from pygame import Rect
from albow.widget import Widget
from albow.controls import Label, Button, Row, Column, Grid, \
	IntEditor, Image
from albow.dialogs import Dialog
from albow.resource import get_image, get_font, get_sound
from albow.image_array import get_image_array
from machine import Machine, MachineNotConfiguredError
from yoffset import YOffset, YOffsetField
from controls import ImageMultichoice, MachineEditor
from part import Part
from game import game

#------------------------------------------------------------------------

class PunchTool(object):
	#  shape   'rect' or 'circle'
	#  _size   (width, height)
	#  purchased  boolean
	
	title = "Punching Tool"
	base_cost = 50
	area_cost = 0.5
	perimeter_cost = 1
	
	purchased = False
	_size = (None, None)
	
	def __init__(self, shape, size):
		self.shape = shape
		self.size = size
	
	def get_size(self):
		if self.shape == 'circle':
			d = self._size[0]
			return (d, d)
		else:
			return self._size
		
	def set_size(self, x):
		self._size = x
	
	size = property(get_size, set_size)
	
	def calculate_cost(self):
		shape = self.shape
		width, height = self.size
		if shape == 'rect':
			area = width * height
			perimeter = 2 * (width + height)
		elif shape == 'circle':
			r = width
			area = pi * r * r
			perimeter = 2 * pi * r
		else:
			raise ValueError("Invalid tool shape")
		return int(self.base_cost +
			self.area_cost * area +
			self.perimeter_cost * perimeter)
	
	cost = property(calculate_cost) 
	
	def check_ready(self):
		if not self.shape or not self.size:
			raise MachineNotConfiguredError

#------------------------------------------------------------------------

shape_images = get_image_array("punch_shape_choices.png", 2, border = 2)
#edge_images = get_image_array("punch_edge_choices.png", (2, 2), border = 2)
edge_images = get_image_array("left_right.png", 2, border = 2)

class PunchEditor(MachineEditor):

	def create_contents(self, machine):
		rows = [
			[Label("Tool Shape"), ToolShapeField(machine)],
			[ToolSizeLabel(machine), ToolSizeField(machine)],
			[Label("X Offset"), XOffsetField(machine)],
			[Label("Y Offset"), YOffsetField(machine)],
			[Label("No. of Strokes"), NumStrokesField(machine)],
			[Label("Advance (A)"), AdvanceField(machine)],
			[Label("Direction"), EdgeField(machine)],
		]
		return Column([
			Grid(rows),
			DiagramView(machine),
		])

def ToolShapeField(machine):
	return ImageMultichoice(images = shape_images, border_width = 1,
		values = ['rect', 'circle'], attribute = 'tool_shape', model = machine)

def ShapeDependentField(machine, f1, f2):
	def is_rect():
		return machine.tool_shape == 'rect'
	def is_circle():
		return machine.tool_shape == 'circle'
	f1.get_visible = is_rect
	f2.get_visible = is_circle
	widget = Widget()
	widget.add([f1, f2])
	widget.shrink_wrap()
	return widget

def ToolSizeLabel(machine):
	lbl1 = Label("Tool Size")
	lbl2 = Label("Tool Diameter")
	return ShapeDependentField(machine, lbl1, lbl2)

def ToolSizeField(machine):
	f1 = ToolWidthHeightField(machine)
	f2 = ToolDiameterField(machine)
	return ShapeDependentField(machine, f1, f2)

def DimensionField(min, max, attr, machine):
	return IntEditor(min = min, max = max, empty = None,
		attribute = attr, model = machine)

def ToolWidthHeightField(machine):
	wf = DimensionField(1, 60, '_tool_width', machine)
	hf = DimensionField(1, 60, '_tool_height', machine)
	return Row([wf, Label("x"), hf])

def ToolDiameterField(machine):
	return DimensionField(1, 60, 'tool_width', machine)

def XOffsetField(machine):
	return DimensionField(-30, 30, 'xoffset', machine)

def NumStrokesField(machine):
	return IntEditor(min = 1, width = "99", empty = None, attribute = 'num_strokes',
		model = machine)

def AdvanceField(machine):
	return DimensionField(1, 60, 'advance', machine)

#class EdgeImages(object):
#
#	indexes = {'rect': 0, 'circle': 1}
#
#	def __init__(self, machine):
#		self.machine = machine
#	
#	def __len__(self):
#		return 2
#	
#	def __getitem__(self, i):
#		s = self.machine.tool_shape or 'rect'
#		j = self.indexes[s]
#		return edge_images[j, i]

def EdgeField(machine):
	return ImageMultichoice(images = edge_images, border_width = 1,
		values = ['left', 'right'], attribute = 'edge', model = machine)

class DiagramView(Image):

	images = get_image_array("punch_diagrams.png", (2, 2), border = 2)
	
	def __init__(self, machine):
		Image.__init__(self, rect = self.images.get_rect())
		self.machine = machine
	
	def get_image(self):
		m = self.machine
		i = m.tool_shape == 'circle'
		j = m.edge == 'right'
		return self.images[i, j]

#------------------------------------------------------------------------

class Punch(YOffset, Machine):
	#  tool_shape       'rect' or 'circle'
	#  tool_width       int
	#  tool_height      int
	#  xoffset          int
	#  yoffset          int
	#  num_strokes      int
	#  advance          int
	#  edge             'left' or 'right'
	#  tool             PunchTool
	#  state            'idle', 'extending', 'retracting'
	#  timeout          int
	#  stroke           int
	#  handling_part    Part
	#  collected_parts  [Part]
	#  produced_parts   [Part]
	#  last_part        Part
	
	title = "Punch"
	icon = get_image("punch_icon.png")
	layer = 1
	image = get_image("punch.png")
	cost = 2000
	editor_class = PunchEditor
	cycle_frames = 5
	resting_frames = 0 # 5
	position_tolerance = 4
	max_extension = 10
	cycle_sound = get_sound("punch.ogg")
	release_sound = get_sound("short_drop.ogg")
	resettable = True
	
	local_bounding_rect = Rect(-32, -96, 64, 128)

	local_covered_rects = [
		[Rect(-32, -64, 64, 32)],
		[local_bounding_rect]
	]

	local_input_rect = Rect(-32, -32, 64, 64)
	local_delivery_point = (0, -96)
	
	tool_image_array = get_image_array("punch_tools.png", 2)
	tool_images = {
		'circle': tool_image_array[0],
		'rect': tool_image_array[1],
	}
	
	tool = None
	tool_shape = None
	tool_width = None
	tool_height = None
	xoffset = None
	num_strokes = 1
	advance = None
	edge = None
	state = 'idle'
	timeout = 0
	stroke = 0
	handling_part = None
	last_part = None
	
	def __init__(self):
		self.collateral_parts = []
		self.produced_parts = []
	
	def __setstate__(self, d):
		if 'xoffset' not in d:
			d['xoffset'] = d.pop('offset')
		self.__dict__.update(d)
		if 'produced_parts' not in d:
			self.produced_parts = []
		if self.tool:
			self.tool_shape = self.tool.shape
			self.tool_width, self.tool_height = self.tool.size
	
	def yoffset_restriction(self):
		if self.tool_shape == 'circle':
			return self.tool_width
		else:
			return self.tool_height
	
	def get_tool_width(self):
		return self.tool_width
	
	def set_tool_width(self, w):
		self.tool_width = w
		self.limit_yoffset()
	
	_tool_width = property(get_tool_width, set_tool_width)
	
	def get_tool_height(self):
		return self.tool_height
	
	def set_tool_height(self, h):
		self.tool_height = h
		self.limit_yoffset()
	
	_tool_height = property(get_tool_height, set_tool_height)
	
	def reset(self, factory):
		self.flush(factory, 'handling_part')
		self.state = 'idle'
		self.timeout = 0
		self.stroke = 0
		self.last_part = None
	
	def check_ready(self):
		if not self.is_configured():
			raise MachineNotConfiguredError
		self.configure_tool()
		self.tool.check_ready()
	
	def is_configured(self):
		if not (self.tool_shape and self.tool_width):
			return False
		if self.tool_shape <> 'circle':
			if not self.tool_height:
				return False
		if not (self.edge and self.xoffset is not None):
			return False
		if self.num_strokes <> 1 and self.advance is None:
			return False
		return True
	
	def find_unpurchased_components(self, result):
		if not self.tool.purchased:
			result.append(self.tool)
	
	def extract_components(self, result):
		tool = self.tool
		if tool:
			result.append(tool)
			self.tool = None

	def configure_tool(self):
		shape = self.tool_shape
		if shape == 'circle':
			size = (self.tool_width, self.tool_width)
		else:
			size = (self.tool_width, self.tool_height)
		def match(tool):
			return tool and tool.shape == shape and tool.size == size
		company = game.company
		if not match(self.tool):
			if self.tool:
				if self.tool.purchased:
					company.put_machine_into_storage(self.tool)
					self.tool = company.get_machine_from_storage(PunchTool, match)
				else:
					self.tool = None
			if not self.tool:
				self.tool = PunchTool(shape, size)

	def begin_frame(self, factory):
		if self.timeout > 0:
			self.timeout -= 1
		else:
			self.change_state(factory)
		del self.collateral_parts[:]
	
	def change_state(self, factory):
		state = self.state
		if state == 'idle':
			pass
		elif state == 'extending':
			self.process_parts(factory)
			self.state = 'retracting'
			self.timeout = self.cycle_frames
		elif state == 'retracting':
			self.deliver_parts(factory)
			self.release_part()
			self.state = 'idle'
			#self.state = 'resting'
			#self.timeout = self.resting_frames
			factory.queue_sound(self.release_sound)
			self.stroke += 1
			if self.stroke >= self.num_strokes:
				self.stroke = 0
		else:
			self.state = 'idle'
		#print "Punch.change_state:", state, "-->", self.state
	
	def handle_part(self, factory, part):
		#print "Punch: Deciding whether to handle part", part ###
		result = False
		if self.state == 'idle' \
				and (self.stroke > 0 or part is not self.last_part) \
				and not self.handling_part \
				and self.part_in_input_rect(part):
			if part is not self.last_part:
				self.stroke = 0
			d = self.part_misalignment(part)
			#print "Punch: misalignment =", d ###
			if abs(d) <= self.position_tolerance:
				#print "Punch: Handling part", part ###
				factory.move_part(part, (-d, 0))
				self.handling_part = part
				part.handler = self
				self.state = 'extending'
				self.timeout = self.cycle_frames
				factory.queue_sound(self.cycle_sound)
				result = True
		if not part.handler:
			#print "Punch: Adding collateral part", part ###
			self.collateral_parts.append(part)
		#print "Punch: Decided whether to handle part:", result ###
		return result

	def part_misalignment(self, part):
		x, y = self.global_origin()
		r = part.global_rect()
		edge = self.edge
		w = self.tool.size[0]
		s = self.stroke_offset()
		if edge == 'left':
			d = r.left - (x - (w // 2) - s)
		else:
			d = r.right - (x + ((w + 1) // 2) + s)
		return d
	
	def stroke_offset(self):
		return self.xoffset + self.stroke * (self.advance or 0)

	def process_parts(self, factory):
		part = self.handling_part
		if part:
			#print "Punch: Processing handled part", part ###
			self.process_part(factory, part)
		for part in self.collateral_parts:
			#print "Punch: Processing collateral part", part ###
			self.process_part(factory, part)
	
	def release_part(self):
		part = self.handling_part
		if part:
			part.handler = None
			self.handling_part = None
			self.last_part = part

	def process_part(self, factory, input_part):
		ox, oy = self.global_origin()
		tool_part = get_tool_part(self.tool)
		tool_pixels = tool_part.pixels
		tool_rect = tool_part.global_rect()
		tool_rect.center = ox, oy + self.yoffset
		
		input_pixels = input_part.pixels
		input_rect = input_part.global_rect()
		
		off = self.stroke_offset()
		if self.edge == 'left':
			tool_rect.left = input_rect.left + off
		else:
			tool_rect.right = input_rect.right - off
		
		output_rect = input_rect.clip(tool_rect)
		ox, oy, ow, oh = output_rect
		if not (ow or oh):
			return
		output_part = Part(output_rect.size)
		output_pixels = output_part.pixels
		
		ix, iy = input_rect.topleft
		tx, ty = tool_rect.topleft
		hrange = xrange(ox, ox + ow)
		vrange = xrange(oy, oy + oh)
		e = (0, 0, 0, 0)
		empty = True
		#print "Punch.process_part:" ###
		#print "...input position =", ix, iy, "size =", input_rect.size ###
		#print "...tool position =", tx, ty, "size =", tool_rect.size ###
		#print "...output position =", ox, oy, "size =", output_rect.size ###
		for y in vrange:
			for x in hrange:
				a = tool_pixels.get_at((x - tx, y - ty))[3]
				if a:
					c = input_pixels.get_at((x - ix, y - iy))
					output_pixels.set_at((x - ox, y - oy), c)
					input_pixels.set_at((x - ix, y - iy), e)
					if c[3]:
						empty = False
		
		if not empty:
			input_part.tag = 'punch waste'
			#print "Punch: Produced", input_part ###
			self.produced_parts.append(output_part)
		
	def deliver_parts(self, factory):
		for part in self.produced_parts:
			self.deliver_part(factory, part)
		del self.produced_parts[:]

	def deliver_part(self, factory, part):
		part.tag = 'punch product'
		#print "Punch: Delivering", part ###
		p = self.global_point(self.local_delivery_point)
		factory.add_part(part, p)
	
	def draw_local(self, surf, t):
		state = self.state
		if state == 'extending' or state == 'retracting':
			phase = self.max_extension * self.timeout // self.cycle_frames
			if state == 'retracting':
				phase = self.max_extension - phase
			image = self.tool_images[self.tool.shape]
			p = (0, -phase)
			t.draw_image(surf, [image], p, ref = 'midbottom')
		p = (-32, -84)
		t.draw_image(surf, [self.image], p)

#------------------------------------------------------------------------

tool_parts = {}

def get_tool_part(tool):
	key = (tool.shape, tool.size)
	part = tool_parts.get(key)
	if not part:
		part = create_tool_part(tool)
		tool_parts[key] = part
	return part

def create_tool_part(tool):
	part = Part(tool.size)
	shape = tool.shape
	width, height = tool.size
	c = (0, 0, 0, 255)
	if shape == 'rect':
		part.pixels.fill(c)
	elif shape == 'circle':
		r = 0.5 * width
		part.fill_float_circle(c, (r, r), r)
	else:
		raise ValueError("Invalid tool shape")
	return part
