################################################################
#
#   Assembly Line - Transverse Saw
#
################################################################

from pygame import Rect
from albow.controls import Label, Button, Row, Column, Grid, IntEditor
from albow.dialogs import Dialog
from albow.resource import get_image, get_font, get_sound
from albow.image_array import get_image_array
from machine import Machine, MachineNotConfiguredError
from controls import ImageMultichoice, MachineEditor
from part import Part
from game import game

#------------------------------------------------------------------------

edge_images = get_image_array("tran_saw_edge_choices.png", 2, border = 2)

class TransverseSawEditor(MachineEditor):

	def create_contents(self, machine):
		rows = [
			[Label("Cut Width"), CutWidthField(machine)],
			[Label("Direction"), EdgeField(machine)],
		]
		return Grid(rows)
		
def CutWidthField(machine):
	return IntEditor(min = 1, max = 59, empty = None,
		attribute = 'cut_width',  model = machine)

def EdgeField(machine):
	return ImageMultichoice(images = edge_images, border_width = 1,
		values = ['left', 'right'], attribute = 'edge', model = machine)

#------------------------------------------------------------------------

class TransverseSaw(Machine):

	title = "Transverse Saw"
	icon = get_image("tran_saw_icon.png", border = 2)
	layer = 1
	frame_image = get_image("tran_saw_frame.png")
	carriage_image = get_image("tran_saw_carriage.png")
	cap_image = get_image("tran_saw_cap.png")
	saw_sound = get_sound("tran_saw.ogg")
	offcut_sound = get_sound("short_drop.ogg")
	editor_class = TransverseSawEditor
	resettable = True

	local_bounding_rect = Rect(-32, -96, 64, 128)

	local_covered_rects = [
		[Rect(-32, -64, 64, 32)],
		[local_bounding_rect]
	]

	local_input_rect = Rect(-32, -32, 64, 64)
	local_offcut_point = (0, -96)
	local_frame_topleft = (-32, -80)
	
	travel_top = -32
	carriage_left = -15
	cap_offset = (3, -6)
	travel = 39
	travel_frames = 10
	max_phase = travel_frames
	
	cut_width = None
	edge = None # 'left' or 'right'
	
	state = 'idle'
	phase = 0
	held_part = None
	
	def __init__(self):
		self.collateral_parts = []
		self.produced_parts = []
	
	def reset(self, factory):
		self.flush(factory, 'held_part')
		del self.collateral_parts[:]
		self.state = 'idle'
		self.phase = 0
	
	def check_ready(self):
		if not self.cut_width or not self.edge:
			raise MachineNotConfiguredError

	def handle_part(self, factory, part):
		if self.state == 'idle' and not self.held_part:
			if self.part_is_candidate(part):
				d = self.part_misalignment(part)
				if abs(d) <= self.position_tolerance:
					factory.move_part(part, (-d, 0))
					self.hold_part(part)
					self.start_cutting(factory)
					return True
		self.collateral_parts.append(part)
		return False
	
	def part_is_candidate(self, part):
		return part.pixels.get_width() > self.cut_width \
			and self.part_in_input_rect(part)

	def part_misalignment(self, part):
		x, y = self.global_origin()
		r = part.global_rect()
		edge = self.edge
		w = self.cut_width
		if edge == 'left':
			d = r.left - (x - w)
		else:
			d = r.right - (x + w)
		return d
	
	def hold_part(self, part):
		self.held_part = part
		part.handler = self
	
	def get_held_part(self):
		part = self.held_part
		self.held_part = None
		return part
	
	def release_part(self):
		part = self.get_held_part()
		if part:
			part.handler = None
		return part
	
	def start_cutting(self, factory):
		factory.queue_sound(self.saw_sound)
		self.state = 'cutting'
	
	def finish_cutting(self, factory):
		part = self.release_part()
		self.cut(factory, part)
		for part in self.collateral_parts:
			self.cut(factory, part)
		del self.collateral_parts[:]
		self.start_returning(factory)
	
	def start_returning(self, factory):
		self.state = 'returning'
	
	def finish_returning(self, factory):
		part = self.release_part()
		if part and part.pixels.get_width() < self.cut_width:
			p = self.global_point(self.local_offcut_point)
			factory.move_part_to(part, p)
			factory.queue_sound(self.offcut_sound)
		self.state = 'idle'
	
	def cut(self, factory, in_part):
		in_rect = in_part.get_rect()
		#print "TransverseSaw: in_rect =", in_rect ###
		in_width, in_height = in_rect.size
		pr_width = self.cut_width
		if pr_width < in_width:
			factory.remove_part(in_part)
			oc_width = in_width - pr_width - 1
			edge = self.edge
			pr_rect = Rect(0, in_rect.top, pr_width, in_height)
			if edge == 'left':
				pr_rect.left = in_rect.left
			else:
				pr_rect.right = in_rect.right
			#print "TransverseSaw: pr_rect =", pr_rect ###
			pr_part = Part(pr_rect.size)
			pr_part.pixels.blit(in_part.pixels, (in_rect.left - pr_rect.left, 0))
			pr_part.tag = "transverse saw product"
			factory.add_part(pr_part, pr_rect.center)
			if oc_width > 0:
				oc_rect = Rect(0, in_rect.top, oc_width, in_height)
				if edge == 'left':
					oc_rect.right = in_rect.right
				else:
					oc_rect.left = in_rect.left
				#print "TransverseSaw: oc_rect =", oc_rect ###
				oc_part = Part(oc_rect.size)
				oc_part.pixels.blit(in_part.pixels, (in_rect.left - oc_rect.left, 0))
				oc_part.tag = "transverse saw offcut"
				factory.add_part(oc_part, oc_rect.center)
				self.hold_part(oc_part)

	def begin_frame(self, factory):
		state = self.state
		if state ==	'cutting':
			if self.phase < self.max_phase:
				self.phase += 1
			else:
				self.finish_cutting(factory)
		elif state == 'returning':
			if self.phase > 0:
				self.phase -= 1
			else:
				self.finish_returning(factory)
		del self.collateral_parts[:]
	
	def draw_local(self, surf, t):
		cx = self.carriage_left
		cy = self.travel_top + self.phase * self.travel // self.max_phase
		t.draw_image(surf, [self.carriage_image], (cx, cy))
		p = self.local_frame_topleft
		t.draw_image(surf, [self.frame_image], p)
		dx, dy = self.cap_offset
		t.draw_image(surf, [self.cap_image], (cx + dx, cy + dy))
