################################################################
#
#   Assembly Line - Material type
#
################################################################

from Numeric import add, innerproduct, swapaxes, array2string, \
	minimum, NewAxis, asarray
from albow.resource import get_font
from imath import iceil
from machine import MachineNotConfiguredError

pixel_cost_matrix = [
	# k   w    y    m    c    r    g    b
	  1,  4,  10,  12,  14,  10,  300,  1000
]
pixel_cost_scale_factor = 0.1

base_cost = 1
area_factor = 1.0
perimeter_factor = 0.02
sheet_cost_scale_factor = 1.0

formula_font = get_font(15, "Vera.ttf")
formula_subscript_font = get_font(10, "Vera.ttf")

#------------------------------------------------------------------------

class MaterialType:

	r = None
	g = None
	b = None
	
	def check_ready(self):
		if self.r is None or self.g is None or self.b is None:
			raise MachineNotConfiguredError
	
	def rgba(self):
		r = self.r
		g = self.g
		b = self.b
		if r is None or g is None or b is None:
			return (0, 0, 0, 0)
		else:
			return (r * 0x11, g * 0x11, b * 0x11, 0xff)
	
	def set_rgba(self, c):
		self.r = c[0] >> 4
		self.g = c[1] >> 4
		self.b = c[2] >> 4
	
	def get_cost(self):
		return pixel_cost(self.rgba())
	
	cost = property(get_cost)

	def sheet_cost(self, (width, height)):
		pc = pixel_cost(self.rgba())
		width = width or 0
		height = height or 0
		area = width * height * 0.001
		perimeter = 2 * (width + height)
		total = base_cost + pc * area_factor * area + perimeter_factor * perimeter
		#print "Material.sheet_cost: area", area, "perimeter", perimeter, "total", total ###
		#print base_cost, "+", pc * area_factor * area, "+", perimeter_factor * perimeter, "=", total ###
		return sheet_cost_scale_factor * total
	
	def formula(self):
		return text_formula(self.rgba())
	
#------------------------------------------------------------------------

def pixel_cost(rgba):
	r, g, b, a = rgba
	if not a:
		return 0
	r >>= 4; g >>= 4; b >>= 4; a >>= 4
	c = m = y = w = 0
	k = 45 - r - g - b
	w = min(r, g, b)
	r -= w; g -= w; b -= w
	c = min(g, b)
	m = min(r, b)
	y = min(r, g)
	if c: g -= c; b -= c
	if m: r -= m; b -= m
	if y: r -= y; g -= y
	assert r >= 0 and g >= 0 and b >= 0 and \
		c >= 0 and m >= 0 and y >= 0 and w >= 0
	total = 0
	for x, p in zip(pixel_cost_matrix, (k, w, y, m, c, r, g, b)):
		total += x * p
	return pixel_cost_scale_factor * total

def pixels_cost(rgb, a):
	ar = add.reduce
	f = a * (1.0 / 255.0)
	rgb = asarray((rgb >> 4), 'f') * f[:,:,NewAxis]
	r = rgb[:,:,0]
	g = rgb[:,:,1]
	b = rgb[:,:,2]
	k = (45.0 * f) - r - g - b
	w = minimum(r, minimum(g, b))
	r -= w; g -= w; b -= w
	c = minimum(g, b)
	m = minimum(r, b)
	y = minimum(r, g)
	g -= c; b -= c
	r -= m; b -= m
	r -= y; g -= y
	va = (k, w, y, m, c, r, g, b)
	v = ar(ar(va, 1), 1)
	total = innerproduct(pixel_cost_matrix, v)
	return pixel_cost_scale_factor * total

def text_formula(color):
	if color[:3] == (0, 0, 0):
		return "K"
	else:
		items = []
		for c, n in zip("RGB", color):
			n >>= 4
			if n:
				items.append("%s%s" % (c, n))
		return ",".join(items)

def draw_formula(surf, color, point, fg, empty = None, ref = 'topleft'):
	x, y = point
	a = formula_font.get_ascent()
	if not color[3]:
		if empty:
			buf = formula_font.render(empty, True, fg)
			surf.blit(buf, (x, y))
	else:
		if color[:3] == (0, 0, 0):
			buf = formula_font.render("K", True, fg)
			r = buf.get_rect()
			setattr(r, ref, (x, y))
			surf.blit(buf, r)
		else:
			for c, n in zip("RGB", color):
				n >>= 4
				if n:
					buf = formula_font.render(c, True, fg)
					r1 = buf.get_rect()
					setattr(r1, ref, (x, y))
					surf.blit(buf, r1)
					buf = formula_subscript_font.render(str(n), True, fg)
					r2 = buf.get_rect()
					r2.midleft = r1.right, r1.top + a
					surf.blit(buf, r2)
					x = r2.right
