################################################################
#
#   Assembly Line - Material composition display and selection
#
################################################################

from pygame import Rect, draw
from albow.widget import Widget
from albow.controls import Label, Row, Column, IntEditor
from albow.dialogs import Dialog
from albow.palette_view import PaletteView
from material import draw_formula
from game import game

#----------------------------------------------------------------------------

class MaterialViewBase(object):

	def draw_in_rect(self, surf, color, frame):
		if color[3]:
			s = frame.height
			r = Rect(frame.topleft, (s, s))
			surf.fill(color, r)
			if frame.width > s:
				draw_formula(surf, color, (r.right + 5, r.centery), self.fg_color,
					ref = 'midleft')

#----------------------------------------------------------------------------

class MaterialView(MaterialViewBase, Widget):
	#  model   MaterialType
	
	default_height = 25
	margin = 3
	
	def __init__(self, material, width = 100, with_formula = True):
		if not with_formula:
			width = self.default_height
		Widget.__init__(self, width = width, height = self.default_height)
		self.material = material

	def draw(self, surf):
		frame = self.get_margin_rect()
		color = self.material.rgba()
		self.draw_in_rect(surf, color, frame)
		
#----------------------------------------------------------------------------

class MaterialField(MaterialView):

	border_width = 1
	
	def __init__(self, product_model, product_attr, material):
		self.product_model = product_model
		self.product_attr = product_attr
		MaterialView.__init__(self, material)

	def mouse_down(self, e):
		dlog = MaterialPopup(self.product_model, self.product_attr, self.material)
		dlog.topleft = self.local_to_global((0, self.height - 1))
		dlog.present(centered = False)

#----------------------------------------------------------------------------

class MaterialPopup(Dialog):

	border_width = 1
	click_outside_response = True
	
	def __init__(self, product_model, product_attr, material):
		from controls import ProductKeySelector
		self.product_model = product_model
		self.product_attr = product_attr
		rgb = RGBField(material)
		products = ProductKeySelector(model = product_model, attribute = product_attr)
		materials = ProductMaterialSelector(
			height = products.height,
			get_product = self.get_product,
			material = material)
		contents = Column([
			rgb,
			Row([products, materials]),
		], align = 'l')
		Dialog.__init__(self, contents)
	
	def get_product(self):
		key = getattr(self.product_model, self.product_attr)
		return game.product_with_key(key)
	
#----------------------------------------------------------------------------

def RGBField(model):
	r = MaterialComponentField("R", "r", model)
	g = MaterialComponentField("G", "g", model)
	b = MaterialComponentField("B", "b", model)
	result = Row([r, g, b])
	r.focus()
	return result

def MaterialComponentField(title, attr, model):
	f = IntEditor(min = 0, max = 15, empty = None, attribute = attr, model = model)
	return Row([Label(title), f])

#----------------------------------------------------------------------------

class ProductMaterialView(MaterialViewBase, PaletteView):

	column_width = 100
	cell_height = 22
	highlight_style = None
	sel_color = (64, 64, 64)
	selectable = False

	def __init__(self, height, get_product):
		self.get_product = get_product
		nrows = height // self.cell_height
		cell_size = (self.column_width, self.cell_height)
		PaletteView.__init__(self, cell_size, nrows, 1, scrolling = True)
	
	def num_items(self):
		product = self.get_product()
		if product:
			return len(product.palette)
		else:
			return 0
	
	def item_is_selected(self, n):
		return False
	
	def draw_item(self, surf, n, rect):
		color = self.get_product().palette[n]
		s = self.sel_width
		d = - 2 * s
		frame = rect.inflate(d, d)
		if self.selectable:
			x, y = frame.topleft
			h = frame.height
			w = h // 2
			if self.item_is_selected(n):
				pts = [
					(x + s, y + s),
					(x + w, y + w),
					(x + s, y + h - s)]
				draw.polygon(surf, self.sel_color, pts)
			frame.left = x + w + 5
			frame.width -= w + 5
		self.draw_in_rect(surf, color, frame)

#----------------------------------------------------------------------------

class ProductMaterialSelector(ProductMaterialView):

	selectable = True
	
	def __init__(self, height, get_product, material):
		self.material = material
		ProductMaterialView.__init__(self, height, get_product)
	
	def material_no(self, n):
		product = self.get_product()
		if product:
			return product.palette[n]
	
	def item_is_selected(self, n):
		c = self.material_no(n)
		m = self.material
		return m and m.rgba() == c
	
	def click_item(self, n, e):
		product = self.get_product()
		if product:
			c = product.palette[n]
			m = self.material
			m.set_rgba(c)
