################################################################
#
#   Assembly Line - Products
#
################################################################

from __future__ import division
import os, sys
from math import sqrt
from albow.resource import resource_path
from part import Part
from timing import period_for_date
from stats import Stats, StatsHolder
from settings import options
from game import game
import config

#----------------------------------------------------------------------------

class Product(object):
	#  key       string
	#  title     string
	#  part      Part
	#  palette   [color]
	#  price     int
	#  description  string
	#  development_cost   int # nominal
	#  months_to_develop  int  # nominal
	
	description = ""
	months_to_develop = 2
	
	def area(self):
		w, h = self.part.pixels.get_size()
		return w * h

	def development_rate(self, budget):
		nominal_time = self.months_to_develop
		nominal_budget = self.development_cost / nominal_time
		funding_ratio = budget / nominal_budget
		rate = sqrt(funding_ratio) / nominal_time
		#print "%s:" % self.title ###
		#print "   nominal_time =", nominal_time ###
		#print "   development cost =", self.development_cost ###
		#print "   research budget =", budget ###
		#print "   nominal budget =", nominal_budget ###
		#print "   funding ratio = %s" % funding_ratio ###
		#print "   research rate =", rate ###
		return rate

	def estimated_development_time(self, history, budget):
		rate = self.development_rate(budget)
		if rate:
			return (1.0 - history.research_done) / rate

#----------------------------------------------------------------------------

class Crate(object):
	#  num_items         int
	#  average_quality   0.0 - 1.0
	#  factory           Factory
	
	factory = None
	
	def __init__(self, num_items, average_quality, factory):
		self.num_items = num_items
		self.average_quality = average_quality
		self.factory = factory
		factory.add_stats(items_produced = num_items)

#----------------------------------------------------------------------------

class ProductStats(Stats):
	#  number_sold             int
	#  number_trashed          int
	#  revenue                 int
	#  freight_cost            int
	#  satisfied_customers     int
	#  dissatisfied_customers  int
	#  sales_missed            int
	#  sales_lost              int
	
	number_sold = 0
	number_trashed = 0
	revenue = 0
	freight_cost = 0
	satisfied_customers = 0
	dissatisfied_customers = 0
	sales_missed = 0
	sales_lost = 0
	
	def calc_percent_satisfied(self):
		sc = self.satisfied_customers
		dc = self.dissatisfied_customers
		tc = sc + dc
		if tc:
			return 100 * sc / tc
	
	percent_satisfied = property(calc_percent_satisfied)
	
#----------------------------------------------------------------------------

class ProductHistory(StatsHolder):
	#  current_price            int
	#  is_for_sale              bool
	#  crates                   [Crate]
	#  stats                    list[period] of ProductStats
	#  total                    ProductStats
	#  potential_market         int
	#  prospective_customers    int
	#  unprospective_customers  int
	#  lost_customers           int
	#  retailer_stock           [Crate]
	#  research_budget          int     dollars per month
	#  research_spending        int
	#  research_done            float   0.0 - 1.0
	
	default_total_market = 10000
	stats_class = ProductStats

	def __init__(self, product):
		StatsHolder.__init__(self)
		self.is_for_sale = True
		self.total_market = self.default_total_market
		self.prospective_customers = 100
		self.unprospective_customers = 0
		self.lost_customers = 0
		self.potential_market = self.total_market - self.prospective_customers
		self.positive_reputation = 0.0
		self.negative_reputation = 0.0
		self.research_budget = 0
		self.research_spending = 0
		self.research_done = 0.0
		self.market_saturation_reported = False
		self.current_price = product.price
		self.crates = []
		self.retailer_stock = []
	
	def __setstate__(self, d):
		stats = d['stats']
		if stats and isinstance(stats[0], tuple):
			stats = [t[1] for t in stats]
			d['stats'] = stats
			del d['stats_by_month']
		self.__dict__.update(d)
		if 'retailer_stock' not in d:
			self.retailer_stock = []
		if 'lost_customers' not in d:
			self.lost_customers = 0
		im = options.dbg_increase_market
		"ProductHistory: im =", repr(im) ###
		if im:
			print "ProductHistory: increasing market by", repr(im) ###
			self.total_market += im
			self.potential_customers += im
	
	def number_in_stock(self):
		tot = 0
		for crate in self.crates:
			tot += crate.num_items
		return tot
	
	def crates_in_stock(self):
		return len(self.crates)
	
	def recent_customer_counts(self):
		sc = 0
		dc = 0
		for s in self.stats[-3:]:
			sc += s.satisfied_customers
			dc += s.dissatisfied_customers
		return sc, dc
	
	def reputation_ratio(self):
		pr = round(self.positive_reputation)
		nr = round(self.negative_reputation)
		tr = pr + nr
		if tr:
			return pr / tr
	
	def biased_reputation_ratio(self):
		r = self.reputation_ratio()
		if r is None:
			r = 1.0
		b = config.reputation_bias
		return b + (1.0 - b) * r

	def add_customer(self, date, satisfied):
		if satisfied:
			name = 'satisfied_customers'
		else:
			name = 'dissatisfied_customers'
		d = {name: 1}
		self.add_stats(date, **d)
		self.prospective_customers -= 1
		if satisfied:
			self.positive_reputation += 1.0
		else:
			self.negative_reputation += 1.0
	
	def lose_customers(self, date, n):
		self.prospective_customers -= n
		self.unprospective_customers += n
		#self.lost_customers += n
		self.add_stats(date, sales_lost = n)
	
	def convert_prospective_to_unprospective_customers(self, n):
		pc = self.prospective_customers - n
		uc = self.unprospective_customers + n
		if pc >= 0 and uc >= 0:
			self.prospective_customers = pc
			self.unprospective_customers = uc
			return True

	def convert_potential_to_prospective_customers(self, n):
		pc = self.prospective_customers + n
		mp = self.potential_market - n
		if pc >= 0 and mp >= 0:
			self.prospective_customers = pc
			self.potential_market = mp
			return True

	def convert_potential_to_unprospective_customers(self, n):
		uc = self.unprospective_customers + n
		mp = self.potential_market - n
		if uc >= 0 and mp >= 0:
			self.unprospective_customers = uc
			self.potential_market = mp
			return True
	
	def bored_customers(self, n):
		self.prospective_customers -= n
		self.potential_market += n
		self.add_stats(game.company.date, sales_missed = n)
	
	def retailer_stock_level(self):
		tot = 0
		for crate in self.retailer_stock:
			tot += crate.num_items
		return tot
	
#----------------------------------------------------------------------------

class ProductParser(object):

	def __init__(self, f, path):
		self.f = f
		self.path = path
		self.lineno = 0
		self.eof = False
		self.next_line()
	
	def next_line(self):
		line = self.f.readline()
		if line:
			self.lineno += 1
			line = line.strip()
			chunks = line.split(None, 1)
			self.keyword = chunks[0]
			if len(chunks) > 1:
				self.data = chunks[1]
			else:
				self.data = ""
		else:
			self.eof = True
	
	def args(self):
		return self.data.split()
	
	def get_args(self, types):
		args = self.args()
		if len(args) <> len(types):
			self.error("Wrong number of args, expected %s got %s" % (
				len(types), len(args)))
		return [t(a) for (t, a) in zip(types, args)]
	
	def expect_data(self, kw):
		self.check_keyword(kw)
		result = self.data
		self.next_line()
		return result
	
	def expect_args(self, kw, types):
		self.check_keyword(kw)
		result = self.get_args(types)
		self.next_line()
		return result
	
	def expect_arg(self, kw, type):
		return self.expect_args(kw, [type])[0]
	
	def check_keyword(self, kw):
		if self.keyword <> kw:
			self.error("Expected '%s'" % kw)
	
	def color(self, name):
		try:
			return self.palette[name]
		except KeyError:
			self.error("Color '%s' not defined" % name)
	
	def read_product(self):
		description = []
		key = self.expect_data("part")
		title = self.expect_data("title")
		price = self.expect_arg("price", int)
		while self.keyword == "desc":
			description.append(self.data)
			self.next_line()
		size = self.expect_args("size", [int, int])
		part = Part(size)
		self.palette = {}
		while not self.eof:
			kw = self.keyword
			if kw == "color":
				name, r, g, b, a = self.get_args([str, int, int, int, int])
				self.palette[name] = (r * 0x11, g * 0x11, b * 0x11, a * 0x11)
				self.next_line()
			elif kw == "rect":
				c, l, t, w, h = self.get_args([str, float, float, float, float])
				part.fill_float_rect(self.color(c), (l, t, w, h))
				self.next_line()
			elif kw == "circle":
				c, x, y, r = self.get_args([str, float, float, float])
				part.fill_float_circle(self.color(c), (x, y), r)
				self.next_line()
			else:
				self.error("Unknown keyword '%s'" % kw)
		product = Product()
		product.key = key
		product.title = title
		product.price = price
		product.development_cost = config.development_cost_factor * price
		product.months_to_develop = product.development_cost / config.suggested_research_budget
		product.description = " ".join(description)
		product.part = part
		product.palette = [v for v in self.palette.values() if v[3]]
		return product
	
	def error(self, mess):
		raise EnvironmentError("%s, line %s: %s" % (
			self.path, self.lineno, mess))

#----------------------------------------------------------------------------

def load_product(path):
	#print "Loading product:", path ###
	f = open(path)
	parser = ProductParser(f, path)
	product = parser.read_product()
	f.close()
	return product

def load_products():
	dir = resource_path("products")
	filenames = os.listdir(dir)
	filenames.sort()
	products = []
	for filename in filenames:
		if filename.endswith(".prd"):
			try:
				path = os.path.join(dir, filename)
				products.append(load_product(path))
			except EnvironmentError, e:
				sys.stderr.write("%s\n" % e)
	return products
