################################################################
#
#   Assembly Line - Base class for 4-period report with
#                   lifetime totals
#
################################################################

from pygame import Rect
from albow.widget import Widget
from albow.resource import get_font
from game import game
from report import ReportView
from timing import period_start_date
from format import format_dollars

class FourPeriodReport(ReportView, Widget):

	header_height = 20
	row_label_width = 150
	column_width = 100
	column_spacing = 10
	num_columns = 5
	row_label_font = get_font(15, "Vera.ttf")
	column_heading_font = get_font(15, "VeraBd.ttf")
	bold_font = get_font(15, "VeraBd.ttf")

	def __init__(self, selection, **kwds):
		nrows = len(self.row_definitions)
		ncols = self.num_columns
		m = self.margin
		hh = self.header_height
		rh = self.predict_font(kwds).get_linesize()
		lw = self.row_label_width
		cw = self.column_width
		cs = self.column_spacing
		width = 2 * m + lw + ncols * (cs + cw)
		height = 2 * m + hh + nrows * rh
		Widget.__init__(self,
			rect = Rect(0, 0, width, height), **kwds)
	
	def data(self):
		holder = self.stats_holder()
		num_periods = self.num_columns - 1
		cp = game.company.current_period()
		p0 = cp - num_periods + 1
		for p in xrange(p0, p0 + num_periods):
			if p >= 0:
				yield self.month_data(holder, p)
			else:
				yield None
		#yield self.total_data(holder, "Last 3 Mths", cp - 2, 3)
		yield self.lifetime_data(holder)
	
	def month_data(self, holder, period):
		d = period_start_date(period)
		heading = d.strftime("%b %Y")
		stats = holder.stats_for_period(period)
		data = self.stats_data(stats)
		return heading, data
	
	def total_data(self, holder, heading, first_period, num_periods):
		stats = holder.stats_for_periods(first_period, num_periods)
		data = self.stats_data(stats)
		return heading, data
	
	def lifetime_data(self, holder):
		heading = "Lifetime"
		stats = holder.total
		data = self.stats_data(stats)
		return heading, data
	
	def stats_data(self, stats):
		data = []
		if stats:
			for defn in self.row_definitions:
				if defn:
					attr = defn[0]
					data.append(getattr(stats, attr))
				else:
					data.append(None)
		return data

	def draw(self, surf):
		self.draw_row_labels(surf)
		for i, info in enumerate(self.data()):
			if info:
				heading, data = info
				self.draw_column(surf, i, heading, data)
	
	def draw_row_labels(self, surf):
		plain_font = self.row_label_font
		bold_font = self.bold_font
		fg = self.fg_color
		h = plain_font.get_linesize()
		x = self.margin
		y = self.margin + self.header_height
		for defn in self.row_definitions:
			if defn:
				text = defn[1]
				if text[0] == "+":
					font = bold_font
				else:
					font = plain_font
				buf = font.render(text[1:], True, fg)
				surf.blit(buf, (x, y))
			y += h
	
	def draw_column(self, surf, i, heading, data):
		m = self.margin
		cw = self.column_width
		cs = self.column_spacing
		fg = self.fg_color
		x = m + self.row_label_width + (i + 1) * (cs + cw)
		y = m
		buf = self.column_heading_font.render(heading, True, fg)
		r = buf.get_rect()
		r.topright = (x, y)
		surf.blit(buf, r)
		if data:
			y += self.header_height
			plain_font = self.row_label_font
			bold_font = self.bold_font
			defns = self.row_definitions
			h = plain_font.get_linesize()
			i = 0
			for cell_data in data:
				if cell_data is not None:
					defn = defns[i]
					if defn[1][0] == "+":
						font = bold_font
					else:
						font = plain_font
					if len(defn) >= 3:
						formatter = defn[2]
					else:
						formatter = format_dollars
					text = formatter(cell_data)
					buf = font.render(text, True, fg)
					r = buf.get_rect()
					r.topright = (x, y)
					surf.blit(buf, r)
				i += 1
				y += h
